#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Compute optimal celestial wcs considering ADR.
"""
import argparse
from astropy.io import fits
from astropy.wcs import WCS
import numpy as np
from reproject.mosaicking import find_optimal_celestial_wcs
import sys

from .ctext import ctext


def find_optimal_celestial_wcs_with_adr(
        input_list,
        extname_adr,
        verbose=False,
):
    """Compute optimal celestial wcs considering ADR.

    Parameters
    ----------
    input_list : str
        Input text file with the name of the FITS files.
    extname_adr : str
        Extension name to where ADR distortion is stored.
    verbose : bool
        If True, display additional information.
    """

    # read input file with list of 3D images to be combined
    with open(input_list) as f:
        file_content = f.read().splitlines()

    # generate input for reproject.mosaicking.find_optimal_celestial_wcs
    list_of_inputs = []
    for fname in file_content:
        if len(fname) > 0:
            if fname[0] not in ['#']:
                print(f'\n* Reading: {fname}')
                with fits.open(fname) as hdul:
                    header3d = hdul[0].header
                    wcs3d = WCS(header3d)
                    if extname_adr is None:
                        wcs2d = wcs3d.celestial
                        list_of_inputs.append(((header3d['NAXIS2'], header3d['NAXIS1']), wcs2d))
                    else:
                        if extname_adr not in hdul:
                            raise ValueError(f"Expected '{extname_adr}' extension not found")
                        if verbose:
                            print(f'Reading ADR correction from {extname_adr} extension')
                        table_adrcross = hdul[extname_adr].data.copy()
                        # ADR at the center of the field of view
                        delta_x_center_ifu = table_adrcross['Delta_x']
                        delta_y_center_ifu = table_adrcross['Delta_y']
                        x_center_ifu, y_center_ifu = wcs3d.celestial.wcs.crpix  # FITS convention
                        center_ifu_coord = wcs3d.celestial.pixel_to_world(
                            x_center_ifu - 1.0,  # Python convention
                            y_center_ifu - 1.0  # Python convention
                        )
                        if verbose:
                            print(f'Center IFU coord: {center_ifu_coord}')
                        x_center_ifu_corrected = x_center_ifu + delta_x_center_ifu
                        y_center_ifu_corrected = y_center_ifu + delta_y_center_ifu
                        # optimal celestial WCS for corrected cube
                        wcs2d_blue = wcs3d.celestial.deepcopy()
                        wcs2d_blue.wcs.crpix = np.array([x_center_ifu_corrected[0], y_center_ifu_corrected[0]])
                        if verbose:
                            print(f"\nwcs2d_blue:\n{wcs2d_blue}")
                        wcs2d_red = wcs3d.celestial.deepcopy()
                        wcs2d_red.wcs.crpix = np.array([x_center_ifu_corrected[-1], y_center_ifu_corrected[-1]])
                        if verbose:
                            print(f"\nwcs2d_red:\n{wcs2d_red}")
                        list_of_inputs.append(((header3d['NAXIS2'], header3d['NAXIS1']), wcs2d_blue))
                        list_of_inputs.append(((header3d['NAXIS2'], header3d['NAXIS1']), wcs2d_red))

    wcs_mosaic2d, shape_mosaic2d = find_optimal_celestial_wcs(list_of_inputs)

    return wcs_mosaic2d, shape_mosaic2d


def main(args=None):

    # parse command-line options
    parser = argparse.ArgumentParser()
    parser.add_argument("input_list", help="TXT file with list of 3D images to be combined", type=str)
    parser.add_argument("--extname_adr",
                        help="Extension name to where ADR distortion is stored. Defaults to None (no ADR distortion)",
                        type=str)
    parser.add_argument("--output_celestial_2d_wcs", help="filename for output celestial 2D WCS", type=str)
    parser.add_argument("--verbose", help="Display intermediate information", action="store_true")
    parser.add_argument("--echo", help="Display full command line", action="store_true")

    args = parser.parse_args(args)

    if len(sys.argv) == 1:
        parser.print_usage()
        raise SystemExit()

    if args.verbose:
        for arg, value in vars(args).items():
            print(ctext(f'{arg}: {value}', faint=True))

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    input_list = args.input_list
    extname_adr = args.extname_adr
    if extname_adr is not None:
        extname_adr = extname_adr.upper()
    verbose = args.verbose
    output_celestial_2d_wcs = args.output_celestial_2d_wcs

    wcs_mosaic2d, shape_mosaic2d = find_optimal_celestial_wcs_with_adr(
        input_list=input_list,
        extname_adr=extname_adr,
        verbose=extname_adr
    )

    if verbose:
        print(f'\n{wcs_mosaic2d=}')
        print(f'\n{shape_mosaic2d=}')

    if output_celestial_2d_wcs is not None:
        header_2d_wcs = wcs_mosaic2d.to_header()
        if extname_adr is not None:
            header_2d_wcs['EXTN_ADR'] = extname_adr
        else:
            header_2d_wcs['EXTN_ADR'] = '<None>'
        hdu = fits.PrimaryHDU(np.zeros(shape_mosaic2d, dtype=np.uint8), header=header_2d_wcs)
        hdu.writeto(output_celestial_2d_wcs, overwrite=True)


if __name__ == "__main__":
    find_optimal_celestial_wcs_with_adr()



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
"""Compute the flux factor for each image based on the median."""
import sys

from matplotlib.colors import LogNorm
from matplotlib.colors import ListedColormap, BoundaryNorm
from matplotlib.patches import Rectangle
import matplotlib.pyplot as plt
import numpy as np

from numina.array.display.plot_hist_step import plot_hist_step
import teareduce as tea

from .remove_isolated_pixels import remove_isolated_pixels


def compute_flux_factor(image3d, median2d, list_flux_factor_regions, _logger, interactive=False, debug=False,
                        nxbins_half=50, nybins_half=50,
                        ymin=0.495, ymax=1.505):
    """Compute the flux factor for each image based on the median.

    Parameters
    ----------
    image3d : 3D numpy array
        The 3D array containing the images to compute the flux factor.
        Note that this array contains the original images after
        subtracting the bias, if any.
    median2d : 2D numpy array
        The median of the input arrays.
    list_flux_factor_regions : list of tea.SliceRegion2D instances
        The regions of interest for the flux factor computation.
    _logger : logging.Logger
        The logger to use for logging.
    interactive : bool, optional
        If True, enable interactive mode for plots (default is False).
    debug : bool, optional
        If True, enable debug mode (default is False).
    nxbins_half : int, optional
        Half the number of bins in the x direction (default is 50).
    nybins_half : int, optional
        Half the number of bins in the y direction (default is 50).
    ymin : float, optional
        Minimum value for the y-axis (default is 0.495).
    ymax : float, optional
        Maximum value for the y-axis (default is 1.505).

    Returns
    -------
    flux_factor : 1D numpy array
        The flux factor for each image.
    """
    naxis3, naxis2, naxis1 = image3d.shape
    if naxis3 > 7:
        _logger.warning("compute_flux_factor: naxis3 > 7, skipping flux factor computation.")
        return np.ones(naxis3, dtype=float)

    naxis2_, naxis1_ = median2d.shape
    if naxis2 != naxis2_ or naxis1 != naxis1_:
        raise ValueError("image3d and median2d must have the same shape in the last two dimensions.")

    # Create a boolean mask for the flux factor regions
    bool_ff_region = np.zeros((naxis2, naxis1), dtype=bool)
    for ff_region in list_flux_factor_regions:
        bool_ff_region[ff_region.python] = True

    if debug:
        # Histograms before applying flux factor
        xmin = np.min(median2d[bool_ff_region])
        xmax = np.max(median2d[bool_ff_region])
        dx = xmax - xmin
        xminh = xmin - dx / 20
        xmaxh = xmax + dx / 20
        bins0 = np.linspace(xmin, xmax, 100)
        h0, edges0 = np.histogram(median2d[bool_ff_region].flatten(), bins=bins0)
        hstep = (edges0[1] - edges0[0])
        fig, ax = plt.subplots()
        for i in range(naxis3):
            xmax_ = np.max(image3d[i][bool_ff_region])
            bins = np.arange(xmin, xmax_ + hstep, hstep)
            h, edges = np.histogram(image3d[i][bool_ff_region].flatten(), bins=bins)
            plot_hist_step(ax, edges, h, color=f'C{i}', label=f'Image {i+1}')
        plot_hist_step(ax, edges0, h0, color='black', label='Median')
        ax.set_xlim(xminh, xmaxh)
        ax.set_xlabel('pixel value')
        ax.set_ylabel('Number of pixels')
        ax.set_title('Before applying flux factor')
        ax.set_yscale('log')
        ax.legend(loc='upper right')
        plt.tight_layout()
        png_filename = 'histogram_before_flux_factor.png'
        _logger.info(f"saving {png_filename}")
        plt.savefig(png_filename, dpi=150)
        if interactive:
            _logger.info("Entering interactive mode (press 'q' to close figure)")
            plt.show()
        plt.close(fig)

    if naxis3 % 2 == 1:
        # Interactive plot showing the image number at the median position
        argsort = np.argsort(image3d, axis=0)

        fig, (ax1, ax2) = plt.subplots(ncols=2, sharex=True, sharey=True, figsize=(12, 5.5))
        aspect_imshow = 'auto'  # 'equal' or 'auto'

        i_comparison_image = 0   # 0 for median2d, 1, 2,... for image3d[comparison_image-1]

        def on_key(event):
            nonlocal img_ax1, img_ax2
            nonlocal i_comparison_image
            nonlocal aspect_imshow
            update_vmin_vmax = False
            # Determine the current region in the plot
            xmin, xmax = ax1.get_xlim()
            ymin, ymax = ax1.get_ylim()
            ixmin = int(round(xmin))
            ixmax = int(round(xmax))
            iymin = int(round(ymin))
            iymax = int(round(ymax))
            if ixmin > ixmax:
                ixmin, ixmax = ixmax, ixmin
            if iymin > iymax:
                iymin, iymax = iymax, iymin
            if ixmin < 1:
                ixmin = 1
            if ixmax > naxis1:
                ixmax = naxis1
            if iymin < 1:
                iymin = 1
            if iymax > naxis2:
                iymax = naxis2
            region2d = tea.SliceRegion2D(f'[{ixmin}:{ixmax},{iymin}:{iymax}]', mode='fits').python
            if event.key == 'a':
                if aspect_imshow == 'equal':
                    aspect_imshow = 'auto'
                else:
                    aspect_imshow = 'equal'
                ax1.set_aspect(aspect_imshow)
                ax2.set_aspect(aspect_imshow)
                ax1.figure.canvas.draw_idle()
                ax2.figure.canvas.draw_idle()
            elif event.key == ',':
                vmin, vmax = np.min(median2d[region2d]), np.max(median2d[region2d])
                update_vmin_vmax = True
            elif event.key == '/':
                vmin, vmax = tea.zscale(median2d[region2d])
                update_vmin_vmax = True
            elif event.key in ['t', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9']:
                i_comparison_image_previous = i_comparison_image
                if event.key == 't':
                    i_comparison_image += 1
                    if i_comparison_image > naxis3:
                        i_comparison_image = 0
                elif event.key == '0':
                    i_comparison_image = 0
                else:
                    i_comparison_image = int(event.key)
                    if i_comparison_image > naxis3:
                        i_comparison_image = i_comparison_image_previous
                if i_comparison_image != i_comparison_image_previous:
                    if i_comparison_image == 0:
                        comparison_image = median2d
                        ax1_title = 'median2d'
                    else:
                        comparison_image = image3d[i_comparison_image - 1]
                        ax1_title = f'single exposure #{i_comparison_image}'
                    _logger.info(f"Displaying {ax1_title} in left panel.")
                    img_ax1.set_data(comparison_image)
                    ax1.set_title(ax1_title)
                    ax1.figure.canvas.draw_idle()
            elif event.key == '?':
                _logger.info("-" * 79)
                _logger.info("Keyboard shortcuts:")
                _logger.info("'h' or 'r' : reset zoom to full image")
                _logger.info("'p' : pan mode")
                _logger.info("'o' : zoom to rectangle")
                _logger.info("'f' : toggle full screen mode")
                _logger.info("'s' : save figure to PNG file")
                _logger.info("." * 79)
                _logger.info("'a' : toggle aspect='equal' / 'aspect='auto' for imshow")
                _logger.info("',' : set vmin and vmax to min and max of the current region")
                _logger.info("'/' : set vmin and vmax using zscale of the current region")
                _logger.info("'t' : cycle through images (left panel)")
                _logger.info("'0-9' : display specific image number in left panel (0 for median)")
                _logger.info("'?' : display this help message")
                _logger.info("'q' : quit interactive mode")
                _logger.info("'x' : exit program")
                _logger.info("-" * 79)
            elif event.key == 'x':
                _logger.info("Exiting program as per user request ('x' key pressed).")
                plt.close(fig)
                sys.exit(0)

            if update_vmin_vmax:
                img_ax1.set_clim(vmin=vmin, vmax=vmax)
                ax1.figure.canvas.draw_idle()

        fig.canvas.mpl_connect('key_press_event', on_key)

        vmin, vmax = tea.zscale(median2d)
        img_ax1, _, _ = tea.imshow(fig, ax1, median2d, ds9mode=True, vmin=vmin, vmax=vmax,
                                   aspect=aspect_imshow, title='median2d')
        if naxis3 == 3:
            color_list = ['tab:red', 'tab:green', 'tab:blue']
        elif naxis3 == 5:
            color_list = ['tab:red', 'tab:orange', 'tab:green', 'tab:blue', 'tab:purple']
        elif naxis3 == 7:
            color_list = ['tab:red', 'tab:orange', 'tab:olive', 'tab:green', 'tab:cyan', 'tab:blue', 'tab:purple']
        else:
            raise ValueError("Cannot define color map for naxis3 odd and > 7.")
        cmap = ListedColormap(color_list)
        bounds = np.arange(0.5, len(color_list) + 1, 1)   # integer limits: -0.5, 0.5, 1.5, ...
        norm = BoundaryNorm(bounds, cmap.N)
        img_ax2, _, cbar2 = tea.imshow(fig, ax2, argsort[naxis3//2] + 1, ds9mode=True,
                                       cmap=cmap, norm=norm,
                                       title='Image number at median position',
                                       cblabel='Image number', aspect=aspect_imshow)
        cbar2.set_ticks(np.arange(1, naxis3 + 1))
        cbar2.ax.yaxis.set_tick_params(length=0)
        ax1.set_xlim(0.5, naxis1 + 0.5)
        ax1.set_ylim(0.5, naxis2 + 0.5)
        for ff_region in list_flux_factor_regions:
            rect = Rectangle((ff_region.fits[0].start - 0.5, ff_region.fits[1].start - 0.5),
                             ff_region.fits[0].stop - ff_region.fits[0].start + 1,
                             ff_region.fits[1].stop - ff_region.fits[1].start + 1,
                             edgecolor='yellow', facecolor='none', linestyle='--', linewidth=1.5)
            ax1.add_patch(rect)
            # ax2.add_patch(rect)
        plt.tight_layout()
        png_filename = 'image_number_at_median_position.png'
        _logger.info(f"saving {png_filename}")
        plt.savefig(png_filename, dpi=150)
        if interactive:
            _logger.info("Entering interactive mode (press 'q' to close figure, 'x' to quit program)")
            plt.show()
        plt.close(fig)

    xmin = np.min(median2d[bool_ff_region])
    xmax = np.max(median2d[bool_ff_region])
    nxbins = 2 * nxbins_half + 1
    nybins = 2 * nybins_half + 1
    xbin_edges = np.linspace(xmin, xmax, nxbins + 1)
    ybin_edges = np.linspace(ymin, ymax, nybins + 1)
    xbin = (xbin_edges[:-1] + xbin_edges[1:])/2
    ybin = (ybin_edges[:-1] + ybin_edges[1:])/2
    extent = [xbin_edges[0], xbin_edges[-1], ybin_edges[0], ybin_edges[-1]]
    cblabel = 'Number of pixels'
    flux_factor = []
    for idata, data in enumerate(image3d):
        # Ratio image / median
        ratio = np.divide(data, median2d,
                          out=np.zeros_like(median2d, dtype=float),
                          where=median2d != 0)

        # Create 2D histogram
        h, edges = np.histogramdd(
            sample=(ratio[bool_ff_region].flatten(), median2d[bool_ff_region].flatten()),
            bins=(ybin_edges, xbin_edges)
        )
        vmin = np.min(h)
        if vmin == 0:
            vmin = 1
        vmax = np.max(h)

        def on_key(event):
            if event.key == 'x':
                _logger.info("Exiting program as per user request ('x' key pressed).")
                plt.close(fig)
                sys.exit(0)

        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12, 5.5))
        fig.canvas.mpl_connect('key_press_event', lambda event: on_key(event))

        tea.imshow(fig, ax1, h, norm=LogNorm(vmin=vmin, vmax=vmax), extent=extent, aspect='auto', cblabel=cblabel)
        ax1.set_xlabel('pixel value')
        ax1.set_ylabel('ratio image/median')
        ax1.set_title(f'Image #{idata+1}')
        hclean = remove_isolated_pixels(h)
        tea.imshow(fig, ax2, hclean, norm=LogNorm(vmin=vmin, vmax=vmax), extent=extent, aspect='auto', cblabel=cblabel)
        ax2.set_xlabel('pixel value')
        ax2.set_ylabel('ratio image/median')
        ax2.set_title(f'Image #{idata+1}')
        xmode = np.zeros(2)
        ymode = np.zeros(2)
        good_splfits = True
        for side, imin, imax in zip((0, 1), (0, nybins_half+1), (nybins_half, nybins)):
            xfit = []
            yfit = []
            for i in range(imin, imax):
                fsum = np.sum(hclean[i, :])
                if fsum > 0:
                    pdensity = hclean[i, :] / fsum
                    perc = (1 - 1 / fsum)
                    p = np.interp(perc, np.cumsum(pdensity), np.arange(nxbins))
                    ax2.plot(xbin[int(p+0.5)], ybin[i], 'x', color=f'C{side}')
                    xfit.append(xbin[int(p+0.5)])
                    yfit.append(ybin[i])
            xfit = np.array(xfit)
            yfit = np.array(yfit)
            try:
                splfit = tea.AdaptiveLSQUnivariateSpline(yfit, xfit, t=2, adaptive=False)
                good_splfit = True
            except Exception as e:
                _logger.warning(f"Could not fit spline to flux factor data for image #{idata+1}, side {side}: {e}")
                good_splfit = False
            if good_splfit:
                ax2.plot(splfit(yfit), yfit, f'C{side}-')
                knots = splfit.get_knots()
                ax2.plot(splfit(knots), knots, f'C{side}o', markersize=4)
                imax = np.argmax(splfit(yfit))
                ymode[side] = yfit[imax]
                xmode[side] = splfit(ymode[side])
                ax2.plot(xmode[side], ymode[side], f'C{side}o', markersize=8)
            if not good_splfit:
                good_splfits = False
                _logger.warning(f"Skipping flux factor computation for image #{idata+1} due to bad spline fit.")
        ax2.set_xlim(ax1.get_xlim())
        ax2.set_ylim(ax1.get_ylim())
        if good_splfits:
            if xmode[0] > xmode[1]:
                imode = 0
            else:
                imode = 1
            ax2.axhline(ymode[imode], color=f'C{imode}', linestyle=':')
            ax2.text(xbin[-5], ymode[imode]+(ybin[-1]-ybin[0])/40, f'{ymode[imode]:.3f}', color=f'C{imode}', ha='right')
            flux_factor.append(ymode[imode])
        else:
            ax2.axhline(1.0, color='k', linestyle=':')
            ax2.text(xbin[-5], 1.01, f'{1.0:.3f}', color='k', ha='right')
            flux_factor.append(1.0)
        plt.tight_layout()

        png_filename = f'flux_factor{idata+1}.png'
        _logger.info(f"saving {png_filename}")
        plt.savefig(png_filename, dpi=150)
        if interactive:
            _logger.info("Entering interactive mode (press 'q' to close figure, 'x' to quit program)")
            plt.show()
        plt.close(fig)

    if len(flux_factor) != naxis3:
        raise ValueError(f"Expected {naxis3} flux factors, but got {len(flux_factor)}.")

    # round the flux factor to 6 decimal places to avoid
    # unnecessary precision when writting to the FITS header
    flux_factor = np.round(flux_factor, decimals=6)

    if debug:
        # Histograms after applying flux factor
        fig, ax = plt.subplots()
        for i in range(naxis3):
            xmax_ = np.max(image3d[i][bool_ff_region])
            bins = np.arange(xmin, xmax_ + hstep, hstep)
            h, edges = np.histogram(image3d[i][bool_ff_region].flatten() / flux_factor[i], bins=bins)
            plot_hist_step(ax, edges, h, color=f'C{i}', label=f'Image {i+1}')
        plot_hist_step(ax, edges0, h0, color='black', label='Median')
        ax.set_xlim(xminh, xmaxh)
        ax.set_xlabel('pixel value')
        ax.set_ylabel('Number of pixels')
        ax.set_title('After applying flux factor')
        ax.set_yscale('log')
        ax.legend(loc='upper right')
        plt.tight_layout()
        png_filename = 'histogram_after_flux_factor.png'
        _logger.info(f"saving {png_filename}")
        plt.savefig(png_filename, dpi=150)
        if interactive:
            _logger.info("Entering interactive mode (press 'q' to close figure)")
            plt.show()
        plt.close(fig)

    return flux_factor



# Copyright 2008-2021 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


from numina.core.oresult import ObservationResult
# import numina.core.instrument.insconf as insconf

from .frame import DataFrameType
from .datatype import DataType


def _obtain_validator_for(instrument, mode_key):
    import numina.drps
    drps = numina.drps.get_system_drps()

    lol = drps.query_by_name(instrument)

    mode = lol.modes[mode_key]
    if mode.validator:
        return mode.validator
    else:
        return lambda mod, obj: True


class ObservationResultType(DataType):
    """The type of ObservationResult."""

    def __init__(self, rawtype=None):
        super(ObservationResultType, self).__init__(ptype=ObservationResult)
        if rawtype:
            self.rawtype = rawtype
        else:
            self.rawtype = DataFrameType

    def validate(self, obj):
        # super(ObservationResultType, self).validate(obj)
        validator = _obtain_validator_for(obj.instrument, obj.mode)
        return validator(self, obj)



# import matplotlib
# matplotlib.use('Qt5Agg')

import matplotlib.pyplot as plt
plt.rcParams.update({'figure.max_open_warning': 0})  # avoid warning


def set_window_geometry(geometry):
    """Set window geometry.

    Parameters
    ==========
    geometry : tuple (4 integers) or None
        x, y, dx, dy values employed to set the Qt backend geometry.

    """

    if geometry is not None:
        x_geom, y_geom, dx_geom, dy_geom = geometry
        mngr = plt.get_current_fig_manager()
        if 'window' in dir(mngr):
            try:
                mngr.window.setGeometry(x_geom, y_geom, dx_geom, dy_geom)
            except AttributeError:
                pass
            else:
                pass



from ..generic import ComponentGeneric, make_component_class, InstrumentGeneric
from ..configorigin import ElementOrigin
from ..property import PropertyEntry, PropertyModLimits


def test_uuid1():

    comp = ComponentGeneric("testcomp")
    assert comp.uuid is None


def test_uuid2():

    comp = ComponentGeneric("testcomp")
    uuid_str = "8e87de0f-2584-4618-9bdb-76c6420f26cf"
    origin = ElementOrigin("testconf", uuid_str)
    comp.set_origin(origin)

    assert str(comp.uuid) == uuid_str
    assert str(comp.origin.uuid) == uuid_str


def test_component_1():

    props = {"p1": PropertyEntry(values=1, depends=[])}
    comp1 = ComponentGeneric.from_component("TEST1", "testcomp1", properties=props)
    assert hasattr(comp1, "p1")
    comp2 = ComponentGeneric("testcomp2")
    assert not hasattr(comp2, "p1")


def test_component_2():
    props = {"p1": PropertyEntry(values=1, depends=[])}

    cls = make_component_class("B", ComponentGeneric, props)

    obj = cls("b")  # noqa
    assert hasattr(obj, "p1")
    assert isinstance(obj, ComponentGeneric)


def test_component_3():
    props = {"p1": PropertyEntry(values=1, depends=[])}

    obj = ComponentGeneric.from_component("B", "b", properties=props)
    assert isinstance(obj, ComponentGeneric)


def test_component_4():
    """Each object has a value of the property"""
    props = {"p1": PropertyModLimits(limits=[-1, 3], default=0)}

    cls = make_component_class("B", ComponentGeneric, props)

    obj1 = cls("b")  # noqa
    obj2 = cls("b")  # noqa

    # defaults before setting any value
    assert obj1.p1 == 0
    assert obj2.p1 == 0

    obj1.p1 = 3
    obj2.p1 = 2

    assert obj1.p1 == 3
    assert obj2.p1 == 2


def test_instrument_1():
    props = {"p1": PropertyEntry(values=1, depends=[])}
    comp1 = InstrumentGeneric.from_component("TEST1", "testcomp1", properties=props)
    assert hasattr(comp1, "p1")
    comp2 = InstrumentGeneric("testcomp2")
    assert not hasattr(comp2, "p1")


def test_instrument_2():
    props = {"p1": PropertyEntry(values=1, depends=[])}

    cls = make_component_class("B", InstrumentGeneric, props)

    obj = cls("b")  # noqa
    assert hasattr(obj, "p1")
    assert isinstance(obj, InstrumentGeneric)


def test_instrument_3():
    props = {"p1": PropertyEntry(values=1, depends=[])}

    obj = InstrumentGeneric.from_component("B", "b", properties=props)
    assert isinstance(obj, InstrumentGeneric)


def test_instrument_4():
    """Each object has a value of the property"""
    props = {"p1": PropertyModLimits(limits=[-1, 3], default=0)}

    cls = make_component_class("B", InstrumentGeneric, props)

    obj1 = cls("b")  # noqa
    obj2 = cls("b")  # noqa

    # defaults before setting any value
    assert obj1.p1 == 0
    assert obj2.p1 == 0

    obj1.p1 = 3
    obj2.p1 = 2

    assert obj1.p1 == 3
    assert obj2.p1 == 2



#
# Copyright 2008-2024 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import unittest
import numpy

import numina.array.combine as c


class MinMaxTestCase(unittest.TestCase):
    """Test case for the minmax rejection method."""

    def setUp(self):
        self.nimages = 10
        self.data = [numpy.ones((2, 2))] * self.nimages

    def testBasic1(self):
        """Test value if points rejected are less than the images."""
        for nmin in range(0, self.nimages):
            for nmax in range(0, self.nimages - nmin):
                out = c.minmax(self.data, nmin=nmin, nmax=nmax)
            for v in out[0].flat:
                self.assertEqual(v, 1)
            for v in out[1].flat:
                self.assertEqual(v, 0)
            for v in out[2].flat:
                self.assertEqual(v, self.nimages - nmin - nmax)

    def testBasic2(self):
        """Test value if points rejected are equal to the images."""
        for nmin in range(0, self.nimages):
            nmax = self.nimages - nmin
            out = c.minmax(self.data, nmin=nmin, nmax=nmax)
            for v in out[0].flat:
                self.assertEqual(v, 0)
            for v in out[1].flat:
                self.assertEqual(v, 0)
            for v in out[2].flat:
                self.assertEqual(v, 0)

    @unittest.skip("requires fixing generic_combine routine")
    def testBasic3(self):
        """Test ValueError is raised if points rejected are more than images"""
        for nmin in range(0, self.nimages):
            nmax = self.nimages - nmin + 1

            self.assertRaises(ValueError, c.minmax, self.data,
                              nmin=nmin, nmax=nmax)



#
# Copyright 2008-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""Default logging configuration for Numina CLI."""


LOGCONF = {
    'version': 1,
    'formatters': {
        'simple': {'format': '%(levelname)s: %(message)s'},
        'state': {'format': '%(asctime)s - %(message)s'},
        'unadorned': {'format': '%(message)s'},
        'detailed': {'format': '%(name)s %(levelname)s %(message)s'},
        'extended': {'format': '%(asctime)s - %(name)s - %(levelname)s - %(message)s'}
    },
    'handlers': {
        'unadorned_console': {
            'class': 'logging.StreamHandler',
            'formatter': 'unadorned',
        },
        'simple_console': {
            'class': 'logging.StreamHandler',
            'formatter': 'simple',
            'level': 'INFO'
        },
        'simple_console_warnings_only': {
            'class': 'logging.StreamHandler',
            'formatter': 'simple',
            'level': 'WARNING'
        },
        'detailed_console': {
            'class': 'logging.StreamHandler',
            'formatter': 'detailed',
            'level': 'INFO'
        },
    },
    'loggers': {
        'numina': {
            'handlers': ['simple_console'],
            'level': 'DEBUG',
            'propagate': False
        }
    },
    'root': {
        'handlers': ['simple_console_warnings_only']
    }
}



#
# Copyright 2016-2024 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Fix points in an image given by a bad pixel mask """

import argparse
import sys

from astropy.io import fits
import numpy

from numina.array._bpm import _process_bpm_intl  # noqa
from numina.tools.arg_file_is_new import arg_file_is_new


def process_bpm(method, arr, mask, hwin=2, wwin=2, fill=0, reuse_values=False, subs=False):
    """

    Parameters
    ----------
    method
    arr
    mask
    hwin
    wwin
    fill
    reuse_values
    subs

    Returns
    -------

    """
    # FIXME: we are not considering variance extension
    # If arr is not in native byte order, the cython extension won't work
    if arr.dtype.byteorder != '=':
        narr = arr.byteswap().newbyteorder()
    else:
        narr = arr
    out = numpy.empty_like(narr, dtype='double')

    # Casting, Cython doesn't support well type bool
    cmask = numpy.where(mask > 0, 1, 0).astype('uint8')

    _, proc = _process_bpm_intl(
        method, narr, cmask, out, hwin=hwin, wwin=wwin, fill=fill, reuse_values=reuse_values)
    if subs:
        return out, proc
    else:
        return out


def process_bpm_median(arr, mask, hwin=2, wwin=2, fill=0, reuse_values=False, subs=False):
    """

    Parameters
    ----------
    arr
    mask
    hwin
    wwin
    fill
    reuse_values
    subs

    Returns
    -------

    """
    import numina.array._combine  # noqa

    method = numina.array._combine.median_method()

    return process_bpm(method, arr, mask, hwin=hwin, wwin=wwin, fill=fill, reuse_values=reuse_values, subs=subs)


def main(args=None):

    # parse command-line options
    parser = argparse.ArgumentParser(
        description='description: apply bad-pixel-mask to image'
    )

    # positional arguments
    parser.add_argument("fitsfile",
                        help="Input FITS file name",
                        type=argparse.FileType('rb'))
    parser.add_argument("--bpm", required=True,
                        help="Bad pixel mask",
                        type=argparse.FileType('rb'))
    parser.add_argument("--outfile", required=True,
                        help="Output FITS file name",
                        type=lambda x: arg_file_is_new(parser, x, mode='wb'))

    # optional arguments
    parser.add_argument("--extnum",
                        help="Extension number in input FITS image "
                             "(default=0)",
                        default=0, type=int)
    parser.add_argument("--extnum_bpm",
                        help="Extension number in bad pixel mask image "
                             "(default=0)",
                        default=0, type=int)
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")

    args = parser.parse_args(args=args)

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    # read input FITS file
    with fits.open(args.fitsfile, mode='readonly') as hdulist_image:
        image2d = hdulist_image[args.extnum].data

        naxis2, naxis1 = image2d.shape

        # read bad pixel mask
        # (mask > 0: masked pixels; mask = 0: unmasked pixel)
        with fits.open(args.bpm) as hdulist_bpm:
            image2d_bpm = hdulist_bpm[args.extnum_bpm].data
            if image2d_bpm.shape != (naxis2, naxis1):
                raise ValueError("NAXIS1, NAXIS2 of FITS image and mask do "
                                 "not match")

            # apply bad pixel mask
            hdulist_image[args.extnum].data = process_bpm_median(
                arr=image2d,
                mask=image2d_bpm,
                reuse_values=True
            )

        # save output FITS file
        hdulist_image.writeto(args.outfile)


if __name__ == "__main__":

    main()



#
# Copyright 2024 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Compute Atmospheric Differential Refraction vs wavelength"""

import argparse
from astropy.table import QTable
import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
import sys

from numina.tools.ctext import ctext


def air_refractive_index_15_760(wave_vacuum):
    """Air refractive index (dry air at sea level)

    Equation (1) from Filippenko (1982): refractive index of dry air
    at sea level (P=760 mm Hg, T=15 degree Celsius).

    Parameters
    ----------
    wave_vacuum : `~astropy.units.Quantity`
        Wavelength (vacuum).

    Returns
    -------
    n_air : float
        Refractive index of dray air.

    """

    wave_vacuum_micron = wave_vacuum.to(u.micron).value

    n_air = 64.328 + 29498.1 / (146 - (1 / wave_vacuum_micron) ** 2) + 255.4 / (41 - (1 / wave_vacuum_micron) ** 2)
    n_air = 1 + n_air * 1E-6

    return n_air


def air_refractive_index(wave_vacuum, temperature, pressure_mm, pressure_water_vapor_mm):
    """Air refractive index (general case)

    Equations (2) and (3) from Filippenko (1982).

    Parameters
    ----------
    wave_vacuum : `~astropy.units.Quantity`
        Wavelength (vacuum).
    temperature : `~astropy.units.Quantity`
        Temperature.
    pressure_mm : float
        Pressure (mm Hg).
    pressure_water_vapor_mm : float
        Water vapour pressure (mm Hg).

    Returns
    -------
    n_air : float
        Refractive index of air.

    """

    wave_vacuum_micron = wave_vacuum.to(u.micron).value
    temperature_value = temperature.to(u.Celsius).value

    n_air = (pressure_mm * (1 + (1.049 - 0.0157 * temperature_value) * 1E-6 * pressure_mm)) / \
            (720.833 * (1 + 0.003661 * temperature_value))
    n_air = 1 + (air_refractive_index_15_760(wave_vacuum) - 1) * n_air

    water_factor = (0.0624 - 0.000680/wave_vacuum_micron**2) / \
                   (1 + 0.003661 * temperature_value) * pressure_water_vapor_mm
    n_air -= water_factor * 1E-6

    return n_air


def compute_adr_wavelength(
        airmass,
        reference_wave_vacuum,
        wave_vacuum,
        temperature=7*u.Celsius,
        pressure_mm=600,
        pressure_water_vapor_mm=8,
        debug=False
):
    """Compute differential atmospheric refraction vs wavelength

    Here we employ the same parameters as Filippenko (1982):
    "At an altitude of ~2 km and a latitude of ~ +/-30, average
    conditions are (Allen 1973) P ~ 600 mm Hg, T = 7 Celsius, and
    Pwater ~ 8 mm Hg."

    Parameters
    ----------
    airmass : float
        Airmass.
    reference_wave_vacuum : `~astropy.units.Quantity`
        Reference wavelength to compute the differential refraction
        correction. This wavelength corresponds to a correction of
        zero.
    wave_vacuum : `~astropy.units.Quantity`
        Array containing `nphotons` simulated photons with the
        spectrum requested in the scene block. These values are
        required to compute the differential refraction correction.
        Note that some values in this array could have been set
        to -1 (e.g., removed photons when applying the atmosphere
        transmission).
    temperature : `~astropy.units.Quantity`
        Temperature.
    pressure_mm : float
        Atmospheric pressure (mm Hg).
    pressure_water_vapor_mm : float
        Water vapor pressure (mm Hg).
    debug : bool
        If True, display additional information.

    Returns
    -------
    differential_refraction : `~astropy.units.Quantity`
        Differential refraction at each simulated wavelength.

    """

    if airmass < 1.0:
        raise ValueError(f'Unexpected {airmass=}')

    # zenith distance
    zenith_distance = np.rad2deg(np.arccos(1/airmass)) * u.deg
    if debug:
        print(ctext(f'{airmass=} --> {zenith_distance=}', faint=True))

    # air refractive index for reference wavelength, at the conditions
    # employed by Filippenko (1982)
    n_air_reference = air_refractive_index(
        wave_vacuum=reference_wave_vacuum,
        temperature=temperature,
        pressure_mm=pressure_mm,
        pressure_water_vapor_mm=pressure_water_vapor_mm
    )
    if debug:
        print(ctext(f'Assuming {temperature=}, {pressure_mm=}, {pressure_water_vapor_mm=}', faint=True))
        print(ctext(f'{reference_wave_vacuum=}', faint=True))
        print(ctext(f'{n_air_reference=}', faint=True))

    # air refractive index for all the simulated wavelengths
    n_air = air_refractive_index(
        wave_vacuum=wave_vacuum,
        temperature=temperature,
        pressure_mm=pressure_mm,
        pressure_water_vapor_mm=pressure_water_vapor_mm
    )

    # refraction (plane-parallel atmosphere)
    factor_arcsec_per_radian = 206264.80624709636
    refraction_reference = (n_air_reference - 1) * np.tan(zenith_distance)
    if debug:
        print(ctext('Refraction at reference wavelength (arcsec): ' +
                    f'{refraction_reference * factor_arcsec_per_radian:+.4f}', faint=True))
    refraction = (n_air - 1) * np.tan(zenith_distance)
    differential_refraction = (refraction - refraction_reference) * factor_arcsec_per_radian * u.arcsec
    if debug:
        # avoid negative wavelengths: those are flagged values corresponding
        # to removed photons (e.g., due to the atmosphere transmission)
        iok = np.argwhere(wave_vacuum > 0 * u.m)
        differential_refraction_ok = differential_refraction[iok]
        simulated_wave_ok = wave_vacuum[iok]
        imin = np.argmin(differential_refraction_ok)
        imax = np.argmax(differential_refraction_ok)
        print(ctext(f'Minimum differential refraction: {differential_refraction_ok[imin][0]:+.4f} ' +
                    f'at wavelength: {simulated_wave_ok[imin][0]}', faint=True))
        print(ctext(f'Maximum differential refraction: {differential_refraction_ok[imax][0]:+.4f} ' +
                    f'at wavelength: {simulated_wave_ok[imax][0]}', faint=True))

    return differential_refraction


def main(args=None):

    # parse command-line options
    parser = argparse.ArgumentParser(description="Compute Atmospheric Differential Refraction")
    parser.add_argument("--airmass", help="Airmass", type=float)
    parser.add_argument("--reference_wave_vacuum", help="Reference wavelength (vacuum)", type=float)
    parser.add_argument("--wave_ini", help="Initial wavelength", type=float)
    parser.add_argument("--wave_end", help="Final wavelength", type=float)
    parser.add_argument("--wave_step", help="Wavelength step", type=float)
    parser.add_argument("--wave_unit", help="Wavelength unit (astropy)", type=str)
    parser.add_argument("--temperature", help="Temperature (Celsius)", type=float, default=7)
    parser.add_argument("--pressure", help="Pressure (mmHg)", type=float, default=600)
    parser.add_argument("--pressure_water_vapor", help="Water vapor pressure (mmHg)", type=float, default=8)
    parser.add_argument("--plots", help="Plot intermediate results", action="store_true")
    parser.add_argument("--ndecimal_wave", help="Number of decimal places in wavelength", type=int, default=3)
    parser.add_argument("--ndecimal_adr", help="Number of decimal places in ADR", type=int, default=3)
    parser.add_argument("--echo", help="Display full command line", action="store_true")
    parser.add_argument("--debug", help="Debug", action="store_true")

    args = parser.parse_args(args=args)

    if len(sys.argv) == 1:
        parser.print_usage()
        raise SystemExit()

    if args.debug:
        for arg, value in vars(args).items():
            print(f'--{arg} {value}')

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    if args.wave_ini is None:
        raise ValueError('You must specify --wave_ini')
    if args.wave_end is None:
        raise ValueError('You must specify --wave_end')
    if args.wave_step is None:
        raise ValueError('You must specify --wave_step')

    if args.wave_unit is None:
        raise ValueError('You must specify --wave_unit')
    else:
        try:
            wave_unit = u.Unit(args.wave_unit)
        except ValueError:
            raise ValueError(f'{args.wave_unit} is not a valid unit')

        if not wave_unit.is_equivalent(u.m):
            raise ValueError(f'{args.wave_unit} is not a valid wavelength unit')

        wave_vacuum = np.arange(
            start=args.wave_ini,
            stop=args.wave_end + args.wave_step / 2,
            step=args.wave_step
        ) * wave_unit

    if args.reference_wave_vacuum is None:
        reference_wave_vacuum = (wave_vacuum[0] + wave_vacuum[-1]) / 2
        print(f'Using reference_wave_vacuum={reference_wave_vacuum}')
    else:
        reference_wave_vacuum = args.reference_wave_vacuum * wave_unit

    differential_refraction = compute_adr_wavelength(
        airmass=args.airmass,
        reference_wave_vacuum=reference_wave_vacuum,
        wave_vacuum=wave_vacuum,
        temperature=args.temperature * u.Celsius,
        pressure_mm=args.pressure,
        pressure_water_vapor_mm=args.pressure_water_vapor,
        debug=args.debug
    )

    result = QTable()
    result['Wavelength'] = wave_vacuum
    result['Wavelength'].info.format = f'.{args.ndecimal_wave}f'
    result['ADR'] = differential_refraction
    result['ADR'].info.format = f'.{args.ndecimal_adr}f'
    result.pprint_all()

    if args.plots:
        fig, ax = plt.subplots()
        ax.plot(wave_vacuum, differential_refraction, '.')
        ax.set_xlabel(f'Wavelength ({wave_unit})')
        ax.set_ylabel(f'Atmospheric Differential Refraction ({differential_refraction.unit})')
        ax.set_title(f'Airmass: {args.airmass}, reference wave: {reference_wave_vacuum}\n' +
                     f'Temperature: {args.temperature} deg Celsius, Pressure: {args.pressure} mmHg\n' +
                     f'Water Vapor Pressure: {args.pressure_water_vapor} mmHg')
        ax.axhline(0, linestyle='--', color='grey')
        ax.axvline(reference_wave_vacuum.value, linestyle=':', color='C1')
        plt.tight_layout()
        plt.show()


if __name__ == "__main__":

    main()



#
# Copyright 2008-2021 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import inspect
import collections.abc

from numina.exceptions import ValidationError
from .base import DataTypeBase
from .typedialect import dialect_info


class DataType(DataTypeBase):
    """Base class for input/output types of recipes.

    """

    def __init__(self, ptype, node_type=None, default=None, **kwds):
        super(DataType, self).__init__(**kwds)
        self.node_type = node_type
        self.internal_type = ptype
        self.internal_dialect = dialect_info(self)
        self.internal_default = default
        self.internal_scalar = True
        self.multi_query = False

    def convert(self, obj):
        """Basic conversion to internal type

        This method is intended to be redefined by subclasses
        """
        return obj

    def convert_in(self, obj):
        """Basic conversion to internal type of inputs.

        This method is intended to be redefined by subclasses
        """
        return self.convert(obj)

    def convert_out(self, obj):
        """Basic conversion to internal type of outputs.

        This method is intended to be redefined by subclasses
        """
        return self.convert(obj)

    def validate(self, obj):
        """Validate convertibility to internal representation

        Returns
        -------
        bool
          True if 'obj' matches the data type

        Raises
        -------
        ValidationError
            If the validation fails

        """
        if not isinstance(obj, self.internal_type):
            raise ValidationError(obj, self.internal_type)
        return True

    def add_dialect_info(self, dialect, tipo):
        key = self.__module__ + '.' + self.__class__.__name__
        result = {'fqn': key, 'python': self.internal_type, 'type': tipo}
        self.internal_dialect[dialect] = result
        return result

    def descriptive_name(self):
        return self.name()


class AutoDataType(DataType):
    """Data type for types that are its own python type"""

    def __init__(self, *args, **kwargs):
        super(AutoDataType, self).__init__(ptype=self.__class__, **kwargs)


class AnyType(DataType):
    """Type representing anything"""

    def __init__(self):
        super(AnyType, self).__init__(ptype=self.__class__)

    def convert(self, obj):
        return obj

    def validate(self, obj):
        return True


class NullType(DataType):
    """Data type for None."""

    def __init__(self):
        super(NullType, self).__init__(type(None))

    def convert(self, obj):
        return None

    def validate(self, obj):
        if obj is not None:
            raise ValidationError
        return True


class PlainPythonType(DataType):
    """Data type for Python basic types."""

    def __init__(self, ref=None, validator=None):
        stype = type(ref)
        if ref is None:
            default = None
        else:
            default = stype()

        if validator is None:
            self.custom_validator = None
        elif callable(validator):
            self.custom_validator = validator
        else:
            raise TypeError('validator must be callable or None')

        super(PlainPythonType, self).__init__(stype, default=default)

    def convert(self, obj):
        pre = self.internal_type(obj)

        if self.custom_validator is not None:
            m = self.custom_validator(pre)
            return m

        return pre

    def __str__(self):
        sclass = type(self).__name__
        return f"{sclass}[{self.internal_type}]"


class ListOfType(DataType):
    """Data type for lists of other types."""

    def __init__(self, ref, default=None, index=0, nmin=None, nmax=None, accept_scalar=False,
                 multi_query=None):
        from numina.core.validator import range_validator
        stype = list
        if inspect.isclass(ref):
            node_type = ref()
        else:
            node_type = ref
        super(ListOfType, self).__init__(
            stype, node_type=node_type, default=default)
        self.internal_scalar = False
        self.index = index
        self.nmin = nmin
        self.nmax = nmax
        self.accept_scalar = accept_scalar
        self.len_validator = range_validator(minval=nmin, maxval=nmax)

        # If multi_query is True, we perform N queries concatenated into a list
        # If multi_query is False, we perform 1 query to obtain directly a list
        if multi_query is None:
            self.multi_query = self.node_type.isproduct()
        else:
            self.multi_query = multi_query

    def convert(self, obj):
        if not isinstance(obj, collections.abc.Iterable):
            if self.accept_scalar:
                obj = [obj]
            else:
                raise TypeError("The object received should be iterable"
                                " or the type modified to accept scalar values")

        result = [self.node_type.convert(o) for o in obj]
        self.len_validator(len(result))
        return result

    def validate(self, obj):
        for o in obj:
            self.node_type.validate(o)
        self.len_validator(len(obj))
        return True

    def _datatype_dump(self, objs, where):
        result = []
        old_dest = where
        for idx, obj in enumerate(objs, start=self.index):
            n_where = f'{old_dest}{idx}'
            res = self.node_type._datatype_dump(obj, n_where)
            result.append(res)
        return result

    def _datatype_load(self, objs):
        if not isinstance(objs, collections.abc.Iterable):
            if self.accept_scalar:
                objs = [objs]
            else:
                raise TypeError("The object received should be iterable"
                                " or the type modified to accept scalar values")
        return [self.node_type._datatype_load(obj) for obj in objs]

    def __str__(self):
        sclass = type(self).__name__
        return f"{sclass}[{self.node_type}]"



#
# Copyright 2024-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

from astropy.io import fits
import matplotlib.pyplot as plt

from .raise_valueerror import raise_ValueError


def display_skycalc(faux_skycalc):
    """
    Display sky radiance and transmission.

    Data generated with the SKYCALC Sky Model Calculator tool
    provided by ESO.
    See https://www.eso.org/observing/etc/doc/skycalc/helpskycalc.html

    Parameters
    ----------
    faux_skycalc : str
        FITS file name with SKYCALC predictions.
    """

    with fits.open(faux_skycalc) as hdul:
        skycalc_header = hdul[1].header
        skycalc_table = hdul[1].data

    if skycalc_header['TTYPE1'] != 'lam':
        raise_ValueError(f"Unexpected TTYPE1: {skycalc_header['TTYPE1']}")
    if skycalc_header['TTYPE2'] != 'flux':
        raise_ValueError(f"Unexpected TTYPE2: {skycalc_header['TTYPE2']}")

    wave = skycalc_table['lam']
    flux = skycalc_table['flux']
    trans = skycalc_table['trans']

    # plot radiance
    fig, ax = plt.subplots()
    ax.plot(wave, flux, '-', linewidth=1)
    cwave_unit = skycalc_header['TUNIT1']
    cflux_unit = skycalc_header['TUNIT2']
    ax.set_xlabel(f'Vacuum Wavelength ({cwave_unit})')
    ax.set_ylabel(f'{cflux_unit}')
    ax.set_title('SKYCALC prediction')
    plt.tight_layout()
    plt.show()

    # plot transmission
    fig, ax = plt.subplots()
    ax.plot(wave, trans, '-', linewidth=1)
    ax.set_xlabel(f'Vacuum Wavelength ({cwave_unit})')
    ax.set_ylabel('Transmission fraction')
    ax.set_title('SKYCALC prediction')
    plt.tight_layout()
    plt.show()



#
# Copyright 2008-2018 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

from .recipes import BaseRecipe  # noqa: F401
from .recipeinout import RecipeInput, RecipeResult  # noqa: F401
from .pipeline import InstrumentDRP, Pipeline  # noqa: F401
# from .instrument.insconf import InstrumentConfiguration
from .pipeline import ObservingMode  # noqa: F401
from .pipelineload import drp_load  # noqa: F401
from .dataholders import Parameter, Requirement  # noqa: F401
from .dataholders import Result, Product  # noqa: F401
from .oresult import ObservationResult  # noqa: F401
from numina.types.product import DataProductType  # noqa: F401
from numina.types.frame import DataFrameType  # noqa: F401
from numina.types.dataframe import DataFrame  # noqa: F401



from numina.drps.tests.pattridrp.instrument.components.wheel import PWheel


def test_wheel1(pattri_ins):
    dev = pattri_ins.get_device("wheel")
    assert isinstance(dev, PWheel)


def test_wheel2(pattri_ins):
    dev = pattri_ins.get_device("wheel")
    assert dev.position == 0
    assert dev.label == "U"


def test_wheel3(pattri_ins):
    dev = pattri_ins.get_device("wheel")
    props = dev.get_properties()
    assert "position" in props
    assert "label" in props
    # selected comes from init_config_info()
    assert "selected" in props


def test_wheel4(pattri_ins):
    dev = pattri_ins.get_device("wheel")
    props = dev.get_property_names()
    assert "position" in props
    assert "label" in props
    # selected comes from init_config_info()
    assert "selected" not in props



#
# Copyright 2010-2020 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


try:
    from functools import singledispatch
except ImportError:
    from pkgutil import simplegeneric as singledispatch

import numpy

from numina.types.dataframe import DataFrame
from numina.types.frame import dump_dataframe
from numina.types.array import dump_numpy_array


@singledispatch
def dump(tag, obj, where):
    """

    Parameters
    ----------
    tag
    obj
    where

    Returns
    -------

    """

    if hasattr(tag, '__numina_dump__'):
        return tag.__numina_dump__(obj, where)

    if hasattr(tag, '_datatype_dump'):
        return tag._datatype_dump(obj, where)

    return obj

# It's not clear if I need to register these three
# functions


@dump.register(list)
def _(tag, obj, where):
    return [dump(tag, o, where) for o in obj]


dump.register(numpy.ndarray, lambda t, o, w: dump_numpy_array(o, w))


dump.register(DataFrame, lambda t, o, w: dump_dataframe(o, w))




import numpy
from setuptools import setup, Extension


def get_extensions():
    numpy_include = numpy.get_include()

    extensions = []
    np_api_min = ("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")
    extra_compile_args_cpp = ["-std=c++17"]

    ext1 = Extension('numina.array._combine',
                     ['src/numina/array/src/combinemodule.cc',
                      'src/numina/array/src/operations.cc',
                      'src/numina/array/src/nu_combine_methods.cc',
                      'src/numina/array/src/nu_combine.cc'
                      ],
                     define_macros=[np_api_min],
                     extra_compile_args=extra_compile_args_cpp,
                     include_dirs=[numpy_include])

    extensions.append(ext1)

    ext3 = Extension('numina.array._nirproc',
                     ['src/numina/array/src/nirproc.pyx'],
                     define_macros=[np_api_min],
                     extra_compile_args=extra_compile_args_cpp,
                     include_dirs=[numpy_include],
                     language='c++'
                     )

    extensions.append(ext3)

    ext4 = Extension('numina.array.trace._traces',
                     ['src/numina/array/trace/traces.pyx',
                      'src/numina/array/trace/Trace.cpp'],
                     define_macros=[np_api_min],
                     extra_compile_args=extra_compile_args_cpp,
                     include_dirs=[numpy_include],
                     language='c++')
    extensions.append(ext4)

    ext5 = Extension('numina.array.trace._extract',
                     ['src/numina/array/trace/extract.pyx'],
                     define_macros=[np_api_min],
                     extra_compile_args=extra_compile_args_cpp,
                     include_dirs=[numpy_include],
                     language='c++')
    extensions.append(ext5)

    ext6 = Extension('numina.array.peaks._kernels',
                     ['src/numina/array/peaks/kernels.pyx'],
                     define_macros=[np_api_min],
                     language='c')
    extensions.append(ext6)

    ext7 = Extension('numina.array._bpm',
                     ['src/numina/array/bpm.pyx'],
                     define_macros=[np_api_min],
                     extra_compile_args=extra_compile_args_cpp,
                     include_dirs=[numpy_include],
                     language='c++')
    extensions.append(ext7)

    ext8 = Extension('numina.array._clippix',
                     ['src/numina/array/clippix.pyx'],
                     define_macros=[np_api_min],
                     include_dirs=[numpy_include],
                     language='c')
    extensions.append(ext8)

    return extensions


if __name__ == '__main__':
    setup(
        ext_modules=get_extensions()
    )




import pytest

import numina.core.pipelineload as loader


@pytest.fixture(scope='module')
def drptest():
    return loader.drp_load('numina.drps.tests', 'drptest4.yaml')


def test_load_link(drptest):
    """Test the loader returns a valid config when 'values' is empty"""

    recipe = drptest.get_recipe_object("image_c")
    assert 'obresult' in recipe.query_options
    assert 'accum_in' in recipe.query_options




import os.path

from ..helpers import WorkEnvironment


def test_work1(tmpdir):
    """Test default definitions"""
    base = 'base'
    basedir = str(tmpdir.dirpath(base))
    data = 'data'
    workdir = 'a'
    resultsdir = 'b'
    work = WorkEnvironment(data, basedir, workdir, resultsdir)
    work.sane_work()

    assert work.workdir == os.path.join(basedir, workdir)
    assert work.basedir == basedir
    assert work.resultsdir == os.path.join(basedir, resultsdir)

    index_base = "index.pkl"
    assert work.index_file == os.path.join(work.workdir, index_base)

    assert os.path.isdir(work.workdir)
    assert os.path.isdir(work.resultsdir)
    assert os.path.isdir(work.basedir)
    assert os.path.isfile(work.index_file)




import argparse
import os.path

from numina.array.display.fileinfo import FileInfo


def subsets_of_fileinfo_from_txt(filename):
    """Returns a dictionary with subsets of FileInfo instances from a TXT file.

    Each subset of files must be preceded by a line:
    @ <number> <label>
    where <number> indicates the number of files in that subset,
    and <label> is a label for that subset. Any additional text
    beyond <label> in the same line is ignored.

    Note that blank lines or lines starting by the hash symbol are
    also ignored. The name of the files comprising each subset will be
    obtained from the first contiguous character string in every
    line (thus, the rest of the line will be discarded).

    Parameters
    ----------
    filename : string
        Name of a TXT file containing a list of FITS files grouped
        in different subsets by the @ symbol.

    Returns
    -------
    dict_of_subsets_of_fileinfo : dictionary
        Dictionary containing as many entries as different subsets
        of files available. Each value of the dictionary is a
        dictionary with a label (sequential number starting at zero)
        and the list of FileInfo instances within subset.

    """

    # check for input file
    if not os.path.isfile(filename):
        raise ValueError("File " + filename + " not found!")

    # read input file
    with open(filename) as f:
        file_content = f.read().splitlines()

    # obtain the different subsets of files
    dict_of_subsets_of_fileinfo = {}
    label = None
    sublist_of_fileinfo = []
    idict = 0
    ifiles = 0
    nfiles = 0
    sublist_finished = True
    for line in file_content:
        if len(line) > 0:
            if line[0] != '#':
                if label is None:
                    if line[0] == "@":
                        nfiles = int(line[1:].split()[0])
                        label = line[1:].split()[1]
                        sublist_of_fileinfo = []
                        ifiles = 0
                        sublist_finished = False
                    else:
                        raise ValueError("Expected @ symbol not found!")
                else:
                    if line[0] == "@":
                        raise ValueError("Unexpected @ symbol found!")
                    tmplist = line.split()
                    tmpfile = tmplist[0]
                    if len(tmplist) > 1:
                        tmpinfo = tmplist[1:]
                    else:
                        tmpinfo = None
                    if not os.path.isfile(tmpfile):
                        raise ValueError("File " + tmpfile + " not found!")
                    sublist_of_fileinfo.append(FileInfo(tmpfile, tmpinfo))
                    ifiles += 1
                    if ifiles == nfiles:
                        dict_of_subsets_of_fileinfo[idict] = {}
                        tmpdict = dict_of_subsets_of_fileinfo[idict]
                        tmpdict['label'] = label
                        tmpdict['list_of_fileinfo'] = sublist_of_fileinfo
                        idict += 1
                        label = None
                        sublist_of_fileinfo = []
                        ifiles = 0
                        sublist_finished = True

    if not sublist_finished:
        raise ValueError("Unexpected end of sublist of files.")

    return dict_of_subsets_of_fileinfo


def main(args=None):

    # parse command-line options
    parser = argparse.ArgumentParser()
    parser.add_argument("txt_file",
                        help="txt file with list of subsets of images")
    args = parser.parse_args(args)

    # execute function
    dict_of_subsets = subsets_of_fileinfo_from_txt(args.txt_file)
    for idict in range(len(dict_of_subsets)):
        tmpdict = dict_of_subsets[idict]
        print('\n>>> Label: ', tmpdict['label'])
        print('>>> List of FileInfo instances:')
        for item in tmpdict['list_of_fileinfo']:
            print(item)


if __name__ == "__main__":

    main()



#
# Copyright 2018-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Frame utils"""

import astropy.io.fits as fits


def copy_img(img):
    """Copy an HDUList"""
    return fits.HDUList([hdu.copy() for hdu in img])



from numina.instrument.generic import ComponentGeneric
from numina.instrument.hwdevice import HWDevice


def test_instrument0(pattri_ins):
    assert isinstance(pattri_ins, ComponentGeneric)
    assert pattri_ins.__class__.__name__ == "InstrumentGeneric_PATTRI"
    assert pattri_ins.name == "PATTRI"


def test_instrument1(pattri_ins):
    base_names = {
        "detector": "Detector_Model1",
        "PSU": "PSU",
        "robot": "RobotPos",
        "wheel": "PWheel",
    }

    for child in base_names:
        obj = pattri_ins.children[child]
        comp_id = child
        assert obj.name == comp_id

        if child in ["PSU", "detector"]:
            assert isinstance(obj, ComponentGeneric)
            assert obj.__class__.__name__ == f"ComponentGeneric_{base_names[child]}"
        if child in ["robot", "wheel"]:
            assert isinstance(obj, HWDevice)
            assert obj.__class__.__name__ == base_names[child]


def test_comp_psu_props(pattri_ins):

    assert pattri_ins.get_property("PSU.box0") == 1
    assert pattri_ins.get_property("PSU.csupos") == 444
    assert pattri_ins.get_property("PSU.box1") == 134
    assert pattri_ins.get_property("PSU.spaces") == [{"mode_a_par": -20}]
    assert pattri_ins.get_property("PSU.other") == [-1]

    psu_dev = pattri_ins.get_device("PSU")
    assert psu_dev.get_property("box0") == 1
    assert psu_dev.get_property("csupos") == 444
    assert psu_dev.get_property("box1") == 134
    assert psu_dev.get_property("spaces") == [{"mode_a_par": -20}]
    assert psu_dev.get_property("other") == [-1]


def test_comp_psu_vals(pattri_ins):
    assert pattri_ins.get_value("PSU.box0") == 1
    assert pattri_ins.get_value("PSU.csupos") == 444
    assert pattri_ins.get_value("PSU.box1", filter="V") == 7126
    assert pattri_ins.get_value("PSU.spaces", insmode="Mode_B") == []
    assert pattri_ins.get_value("PSU.other", tag0="tag0_A", tag1="tag1_Y") == [-2]


def test_comp_psu_vals2(pattri_ins):
    dev = pattri_ins.get_device("PSU")
    assert dev._internal_state == {
        "filter": "B",
        "insmode": "Mode_A",
        "tag0": "tag0_A",
        "tag1": "tag1_X",
    }


def test_comp_psu_comps(pattri_ins):
    assert pattri_ins.get_property("PSU.BarL_1.const1") == 444
    assert pattri_ins.get_property("PSU.BarL_2.const1") == 444
    assert pattri_ins.get_property("PSU.BarR_1.const1") == 444
    assert pattri_ins.get_property("PSU.BarR_2.const1") == 444


def test_comp_robo_comps(pattri_ins):
    assert pattri_ins.get_property("robot.arm_1.angle") == 0
    assert pattri_ins.get_property("robot.arm_1.angle") == 0
    assert pattri_ins.get_property("robot.arm_5.active") is True
    pattri_ins.get_device("robot.arm_4").active = False
    assert pattri_ins.get_property("robot.arm_4.active") is False
    dev = pattri_ins.get_device("robot.arm_4")
    dev.get_property("active")


def test_wheel(pattri_ins):
    dev = pattri_ins.get_device("wheel")
    assert dev.get_property("position") == 0
    dev.turn()
    assert dev.get_property("position") == 1
    assert dev.get_property("label") == "B"
    pattri_ins.get_device("wheel").position = 3
    assert dev.get_property("position") == 3
    assert dev.get_property("label") == "R"
    assert dev._capacity == 5


def test_configure(pattri_ins, pattri_header2):
    pattri_ins.configure_with_header(pattri_header2)
    assert pattri_ins.get_device("wheel").position == 3
    assert pattri_ins.get_property("wheel.position") == 3
    assert pattri_ins.get_property("PSU.box1") == 803
    assert pattri_ins.get_property("PSU.spaces") == [{"mode_a_par": -20}]


def test_configure2(pattri_ins, pattri_header2):
    pattri_ins.configure_with_header(pattri_header2)
    dev = pattri_ins.get_device("PSU")
    assert dev.other == [-1]
    assert pattri_ins.get_property("PSU.other") == [-1]


def test_configure3(pattri_ins, pattri_header):
    pattri_ins.configure_with_header(pattri_header)
    assert pattri_ins.get_device("wheel").label == "R"
    assert pattri_ins.get_property("wheel.position") == 3


def test_configure4(pattri_ins, pattri_header):
    pattri_ins.configure_with_header(pattri_header)
    assert pattri_ins.depends_on() == {"filter", "insmode", "key2", "tag0", "tag1"}


def test_property5(pattri_ins):
    for u in sorted(pattri_ins._property_names):
        print(u)
    print("---")
    for u in sorted(pattri_ins._property_configurable):
        print(u)

    print("--")
    for u in pattri_ins.get_property_names():
        print(u)
    assert pattri_ins.get_property_names() == pattri_ins._property_configurable



#
# Copyright 2016-2018 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""Fit offset and rotation"""

import numpy
import numpy.linalg as linalg


def fit_offset_and_rotation(coords0, coords1):
    """Fit a rotation and a translation between two sets points.

    Fit a rotation matrix and a translation between two matched sets
    consisting of M N-dimensional points

    Parameters
    ----------
    coords0 : (M, N) array_like
    coords1 : (M, N) array_lke

    Returns
    -------
    offset : (N, ) array_like
    rotation : (N, N) array_like

    Notes
    ------
    Fit offset and rotation using Kabsch's algorithm [1]_ [2]_

    .. [1] Kabsch algorithm: https://en.wikipedia.org/wiki/Kabsch_algorithm

    .. [2] Also here: http://nghiaho.com/?page_id=671

    """

    coords0 = numpy.asarray(coords0)
    coords1 = numpy.asarray(coords1)

    cp = coords0.mean(axis=0)
    cq = coords1.mean(axis=0)

    p0 = coords0 - cp
    q0 = coords1 - cq

    crossvar = numpy.dot(numpy.transpose(p0), q0)

    u, _, vt = linalg.svd(crossvar)

    d = linalg.det(u) * linalg.det(vt)

    if d < 0:
        u[:, -1] = -u[:, -1]

    rot = numpy.transpose(numpy.dot(u, vt))
    # Operation is
    # B - B0 = R(A - A0)
    # So off is B0 -R * A0
    # The inverse operation is
    # A - A0 = R* (B- B0)
    # So inverse off* is A - R* B0
    # where R* = transpose(R)
    #  R * off* = -off
    off = -numpy.dot(rot, cp) + cq
    return off, rot



#
# Copyright 2008-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import os


def arg_file_is_new(parser, arg, mode='w'):
    """Auxiliary function to give an error if the file already exists.

    Parameters
    ----------
    parser : parser object
        Instance of argparse.ArgumentParser()
    arg : string
        File name.
    mode : string
        Optional string that specifies the mode in which the file is
        opened.

    Returns
    -------
    handler : file object
        Open file handle.

    """
    if os.path.exists(arg):
        parser.error(
            f"\nThe file \"{arg}\"\nalready exists and cannot be overwritten!")
    else:
        # return an open file handle
        handler = open(arg, mode=mode)
        return handler



#
# Copyright 2024-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

from astropy.io import fits
from astropy.units import Quantity
import astropy.units as u
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np

from .define_3d_wcs import wcs_to_header_using_cd_keywords


def generate_image2d_method0_ifu(
        wcs3d,
        header_keys,
        noversampling_whitelight,
        simulated_x_ifu_all,
        simulated_y_ifu_all,
        prefix_intermediate_fits,
        instname,
        subtitle,
        scene,
        plots
):
    """Compute image2d IFU (white image), method0

    The image is calculated by performing a two-dimensional histogram
    over the X, Y coordinates of the simulated photons in the IFU,
    regardless of the wavelength assigned to each photon.

    The result is saved as a FITS file.

    Parameters
    ----------
    wcs3d : `~astropy.wcs.wcs.WCS`
        WCS of the data cube.
    header_keys : `~astropy.io.fits.header.Header`
        FITS header with additional keywords to be merged together with
        the WCS information.
    noversampling_whitelight : int
        Oversampling factor (integer number) to generate the method0
        white image.
    simulated_x_ifu_all : `~astropy.units.Quantity`
        Simulated X coordinates of the photons in the IFU.
    simulated_y_ifu_all : `~astropy.units.Quantity`
        Simulated Y coordinates of the photons in the IFU.
    prefix_intermediate_fits : str
        Prefix for output intermediate FITS files. If the length of
        this string is 0, no output is generated.
    instname : str or None
        Instrument name.
    subtitle : str or None
        Plot subtitle.
    scene : str
        YAML scene file name.
    plots : bool
        If True, plot intermediate results.

    """

    # select the 2D spatial info of the 3D WCS
    wcs2d = wcs3d.sub(axes=[1, 2])

    naxis1_ifu_oversampled = Quantity(
        value=wcs2d.array_shape[1] * noversampling_whitelight,
        unit=u.pix,
        dtype=int
    )
    naxis2_ifu_oversampled = Quantity(
        value=wcs2d.array_shape[0] * noversampling_whitelight,
        unit=u.pix,
        dtype=int
    )

    bins_x_ifu_oversampled = (0.5 + np.arange(naxis1_ifu_oversampled.value + 1)) * u.pix
    bins_y_ifu_oversampled = (0.5 + np.arange(naxis2_ifu_oversampled.value + 1)) * u.pix

    crpix1_orig, crpix2_orig = wcs2d.wcs.crpix
    crpix1_oversampled = (naxis1_ifu_oversampled.value + 1) / 2
    crpix2_oversampled = (naxis2_ifu_oversampled.value + 1) / 2

    wcs2d.wcs.crpix = crpix1_oversampled, crpix2_oversampled

    # (important: reverse X <-> Y)
    image2d_method0_ifu, xedges, yedges = np.histogram2d(
        x=(simulated_y_ifu_all.value - crpix2_orig) * noversampling_whitelight + crpix2_oversampled,
        y=(simulated_x_ifu_all.value - crpix1_orig) * noversampling_whitelight + crpix1_oversampled,
        bins=(bins_y_ifu_oversampled.value, bins_x_ifu_oversampled.value)
    )

    wcs2d.wcs.cd /= noversampling_whitelight

    # save FITS file
    if len(prefix_intermediate_fits) > 0:
        # use float to avoid saturation problem
        hdu = fits.PrimaryHDU(image2d_method0_ifu.astype(np.float32))
        pos0 = len(hdu.header) - 1
        hdu.header.extend(wcs_to_header_using_cd_keywords(wcs2d), update=True)
        hdu.header.update(header_keys)
        hdu.header.insert(
            pos0, ('COMMENT', "FITS (Flexible Image Transport System) format is defined in 'Astronomy"))
        hdu.header.insert(
            pos0 + 1, ('COMMENT', "and Astrophysics', volume 376, page 359; bibcode: 2001A&A...376..359H"))
        hdul = fits.HDUList([hdu])
        outfile = f'{prefix_intermediate_fits}_ifu_white2D_method0_os{noversampling_whitelight:d}.fits'
        print(f'Saving file: {outfile}')
        hdul.writeto(f'{outfile}', overwrite='yes')

    # display result
    if plots:
        fig, ax = plt.subplots(figsize=(6.4, 6.4))
        img = ax.imshow(image2d_method0_ifu, origin='lower', interpolation='None')
        ax.set_xlabel('X axis (array index)  [parallel to the slices]')
        ax.set_ylabel('Y axis (array index)  [perpendicular to the slices]')
        if instname is not None:
            title = f'{instname} '
        else:
            title = ''
        title += f'IFU image, method0 (oversampling={noversampling_whitelight})'
        if subtitle is not None:
            title += f'\n{subtitle}'
        title += f'\nscene: {scene}'
        ax.set_title(title)
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        fig.colorbar(img, cax=cax, label='Number of photons')
        plt.tight_layout()
        plt.show()



#
# Copyright 2015 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""Tests for RecipeResult"""

from ..dataholders import Result
from ..recipeinout import RecipeResult


def create_result_class():

    class BB(RecipeResult):
        prod1 = Result(int, 'something1')
        prod2 = Result(int, 'something2')

        def somefun(self):
            pass

    return BB


def test_class_construction():

    BB = create_result_class()

    assert issubclass(BB, RecipeResult)


def test_class_desc_access():

    BB = create_result_class()

    assert BB.prod1 is getattr(BB, 'prod1')
    assert BB.prod2 is getattr(BB, 'prod2')

    assert isinstance(BB.prod1, Result)
    assert isinstance(BB.prod2, Result)


def test_class_desc_set():

    BB = create_result_class()

    BB.prod3 = Result(3, 'something3')

    assert BB.prod3 is getattr(BB, 'prod3')
    assert isinstance(BB.prod3, Result)
    assert BB.prod3 is BB.stored()['prod3']
    assert BB.prod3.dest == 'prod3'


def test_class_desc_stored():

    BB = create_result_class()

    stored = BB.stored()

    assert BB.prod1 is stored['prod1']
    assert BB.prod2 is stored['prod2']


def test_ins_desc_access():

    BB = create_result_class()

    bb = BB(prod1=80)

    values = {'prod2': int(), 'prod1': 80}

    for key, val in values.items():
        ival = getattr(bb, key)
        assert val == ival

    # These values are not stored in __dict__
    for key in values:
        assert key not in bb.__dict__


def test_ins_attr_access():

    BB = create_result_class()

    bb = BB(prod1=80)

    values = {'prod2': int(), 'prod1': 80}

    for key, val in bb.attrs().items():
        assert val == values[key]

    # These values are not stored in __dict__
    for key in values:
        assert key not in bb.__dict__


def test_class_nondesc_access():

    BB = create_result_class()
    bb = BB(prod1=80)

    values = {'prod2': int(), 'prod1': 80}

    # If we insert a new attribute,it goes to __dict__
    bb.otherattr = 100

    for key, val in bb.attrs().items():
        assert val == values[key]

    assert 'otherattr' in bb.__dict__



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Compute pixel_to_world.
"""
import argparse

from astropy.io import fits
from astropy.wcs import WCS
import sys

from .ctext import ctext


def pixel_to_world(inputfile, pixel, extnum, verbose=False):
    """Compute world_to_pixel.

    Parameters
    ----------
    inputfile : str, file-like or `pathlib.Path`
        Input FITS filename.
    pixel : str or None
        WCS pixel coordinate.
    extnum : int
        Extension number to read the WCS from.
    verbose : bool
        It True, display intermediate information
    """

    with fits.open(inputfile) as hdul:
        if verbose:
            print(hdul.info())
        if extnum > len(hdul):
            raise ValueError(f"Extension number {extnum} exceeds {len(hdul)}")
        header = hdul[extnum].header

    wcs = WCS(header)
    if verbose:
        print(wcs)
    naxis = wcs.naxis

    if pixel is None or pixel == "":
        pixel = [1] * naxis
    else:
        pixel = [float(item) - 1 for item in pixel.split(',')]   # Python criterion for next function

    result = wcs.pixel_to_world(*pixel)

    print(f'\nComputing world coordinates for pixel {pixel}:')

    if isinstance(result, list):
        print(result[0])
        print(result[1])
    else:
        print(result)


def main(args=None):
    """
    Usage example:
    $ numina-pixel_to_world file.fits --pixel '1,1'

    The pixel coordinates are read as a string. The quote or double
    quote symbol is not necessary if the numbers are given without
    blank spaces.
    """
    parser = argparse.ArgumentParser(description="Convert pixel to world coordinates.")
    parser.add_argument("inputfile", help="Input FITS file", type=str)
    parser.add_argument("--pixel", help="WCS pixel coordinate (comma separated values)", type=str,
                        default=None)
    parser.add_argument("--extnum", help="Extension number (default 0=PRIMARY)", type=int, default=0)
    parser.add_argument("--verbose", help="Display intermediate information", action="store_true")
    parser.add_argument("--echo", help="Display full command line", action="store_true")
    if args is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(args)

    if len(sys.argv) == 1:
        parser.print_usage()
        raise SystemExit()

    if args.verbose:
        for arg, value in vars(args).items():
            print(ctext(f'{arg}: {value}', faint=True))

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    extnum = args.extnum
    if extnum < 0:
        raise ValueError('extnum must be >= 0')

    pixel_to_world(
        inputfile=args.inputfile,
        pixel=args.pixel,
        extnum=extnum,
        verbose=args.verbose
    )


if __name__ == '__main__':
    main()



#
# Copyright 2016-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""Unit test for types"""

from numina.types.datatype import DataType
from ..dataholders import Result
from ..recipeinout import RecipeResult


def test_product_out():

    class AVal(object):
        def __init__(self, val):
            self.val = val

    class A(DataType):

        def __init__(self):
            super(A, self).__init__(ptype=AVal)

        def convert(self, obj):
            return AVal(val=1)

        def convert_out(self, obj):
            return AVal(val=2)

    class RR(RecipeResult):
        prod = Result(A)

    rr = RR(prod=100)

    assert rr.prod.val == 2



#
# Copyright 2016-2021 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

#cython: language_level=3

from cpython.mem cimport PyMem_Malloc, PyMem_Free
from cpython.pycapsule cimport PyCapsule_IsValid
from cpython.pycapsule cimport PyCapsule_GetPointer
from cpython.pycapsule cimport PyCapsule_GetContext
from cython.operator cimport dereference as deref

cimport cython

import numpy as np
cimport numpy as np


ctypedef fused image_t:
    np.float64_t
    np.float32_t
    np.int64_t
    np.int32_t
    np.int16_t
    np.int8_t
    np.uint64_t
    np.uint32_t
    np.uint16_t
    np.uint8_t

# FIXME: this definition should come from the headers in src
ctypedef int (*CombineFunc)(double*, double*, size_t, double*[3], void*)

@cython.boundscheck(False)
@cython.wraparound(False)
def _process_bpm_intl(object method, image_t[:,:] arr, np.uint8_t[:,:] badpix, np.double_t[:,:] res,
                      size_t hwin=2, size_t wwin=2, double fill=0.0, reuse_values=False):
    '''Loop over the mask over a window, border is filled with 'fill' value.'''
    cdef:
        size_t xr = arr.shape[1]
        size_t yr = arr.shape[0]
        size_t ii, x, y, xx, yy
        size_t wsize = (2*hwin+1) * (2*wwin+1)
        int status
        size_t bsize = 0
        double *buff[2]
        double* pvalues[3]
        CombineFunc function = NULL
        void *vdata = NULL


    if not PyCapsule_IsValid(method, "numina.cmethod"):
        raise TypeError("parameter is not a valid capsule")

    function = <CombineFunc>PyCapsule_GetPointer(method, "numina.cmethod")
    vdata = PyCapsule_GetContext(method)

    processed = np.zeros_like(badpix, dtype='uint8')
    cdef np.uint8_t[:, ::1] processed_view = processed

    try:
        # Init memory
        for ii in range(2):
            buff[ii] = <double *> PyMem_Malloc(wsize * sizeof(double))
            if not buff[ii]:
                raise MemoryError()

        for ii in range(3):
            pvalues[ii] = <double *> PyMem_Malloc(sizeof(double))
            if not pvalues[ii]:
                raise MemoryError()

        for x in range(xr):
            for y in range(yr):
                bp = badpix[y, x]

                if bp == 0:
                    # Skip over good pixels
                    res[y, x] = <np.double_t>arr[y, x]
                    processed_view[y, x] = 1
                else:
                    # For bad pixels, use a window
                    for xx in range(x-wwin, x+wwin+1):
                        for yy in range(y-hwin, y+hwin+1):
                            if xx < 0 or yy < 0 or xx >= xr or yy >= yr:
                                buff[0][bsize] = fill
                                buff[1][bsize] = 1.0
                                bsize += 1
                            elif badpix[yy, xx] == 0:
                                buff[0][bsize] = arr[yy, xx]
                                buff[1][bsize] = 1.0
                                bsize += 1
                            elif reuse_values and processed_view[yy, xx] == 1:
                                buff[0][bsize] = res[yy, xx]
                                buff[1][bsize] = 1.0
                                bsize += 1
                            else:
                                continue

                    # Compute values
                    if bsize > 0:
                        status = function(buff[0], buff[1], bsize, pvalues, vdata)
                        res[y,x] = <np.double_t>deref(pvalues[0])
                        processed_view[y, x] = 1
                    else:
                        res[y,x] = fill
                    # reset buffers
                    bsize = 0
        return res, processed
    finally:
        # dealocate memory on exit
        for ii in range(2):
            PyMem_Free(buff[ii])
        for ii in range(3):
            PyMem_Free(pvalues[ii])



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


import numpy

from ._extract import extract_simple_intl
from .traces import axis_to_dispaxis


class Aperture(object):
    """Spectroscopic aperture."""

    def __init__(self, bbox, borders, axis=0, id=None):
        self.bbox = bbox
        self.borders = borders
        self.id = id
        self.axis = axis
        self.dispaxis = axis_to_dispaxis(axis)


def extract_simple_rss(arr, borders, axis=0, out=None):

    # If arr is not in native byte order, the C-extension won't work
    if arr.dtype.byteorder != '=':
        arr2 = arr.byteswap().newbyteorder()
    else:
        arr2 = arr

    if axis == 0:
        arr3 = arr2
    elif axis == 1:
        arr3 = arr2.T
    else:
        raise ValueError("'axis' must be 0 or 1")

    if out is None:
        out = numpy.zeros((len(borders), arr3.shape[1]), dtype='float')

    xx = numpy.arange(arr3.shape[1])

    # Borders contains a list of function objects
    for idx, (b1, b2) in enumerate(borders):
        bb1 = b1(xx)
        bb1[bb1 < -0.5] = -0.5
        bb2 = b2(xx)
        bb2[bb2 > arr3.shape[0] - 0.5] = arr3.shape[0] - 0.5
        extract_simple_intl(arr3, xx, bb1, bb2, out[idx])
    return out


def extract_simple_rss_apers(arr, apers, axis=0, out=None):

    # If arr is not in native byte order, the C-extension won't work
    if arr.dtype.byteorder != '=':
        arr2 = arr.byteswap().newbyteorder()
    else:
        arr2 = arr

    if axis == 0:
        arr3 = arr2
        i1, i2 = 0, 2
    elif axis == 1:
        arr3 = arr2.T
        i1, i2 = 2, 4
    else:
        raise ValueError("'axis' must be 0 or 1")

    if out is None:
        out = numpy.zeros((len(apers), arr3.shape[1]), dtype='float')

    for idx, aper in enumerate(apers):
        if aper.axis != axis:
            raise ValueError("array 'axis' and aperture 'axis' are different")

        u1, u2 = aper.bbox[i1:i2]
        uu = numpy.arange(u1, u2 + 1)

        # Borders contains a list of function objects
        b1, b2 = aper.borders
        bb1 = b1(uu)
        bb1[bb1 < -0.5] = -0.5
        bb2 = b2(uu)
        bb2[bb2 > arr3.shape[0] - 0.5] = arr3.shape[0] - 0.5
        extract_simple_intl(arr3, uu, bb1, bb2, out[idx])

    return out



#
# Copyright 2024-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

from astropy import wcs
from astropy.io import fits
import astropy.units as u
import numpy as np

from .define_3d_wcs import get_wvparam_from_wcs3d
from .define_3d_wcs import wcs_to_header_using_cd_keywords


def save_image2d_rss(
        wcs3d,
        header_keys,
        image2d_rss,
        method,
        prefix_intermediate_fits,
        bitpix
):
    """Save the RSS image.

    Parameters
    ----------
    wcs3d : `~astropy.wcs.wcs.WCS`
        WCS of the data cube.
    header_keys : `~astropy.io.fits.header.Header`
        FITS header with additional keywords to be merged together with
        the WCS information.
    image2d_rss : `~numpy.ndarray`
        2D array containing the RSS image.
    method : int
        Integer indicating the method: 0 or 1.
    prefix_intermediate_fits : str
        Prefix for output intermediate FITS files. If the length of
        this string is 0, no output is generated.
    bitpix : int
        BITPIX value for the FITS file.
    """

    if len(prefix_intermediate_fits) > 0:
        # ------------------------------------------------
        # 1) spectroscopic 2D image with contiguous slices
        # ------------------------------------------------
        # ToDo: compute properly the parameters corresponding to the spatial axis
        # Note that using: wcs2d = wcs3d.sub(axes=[0, 1])
        # selecting the 1D spectral and one of the 1D spatial info of the 3D WCS
        # does not work:
        # "astropy.wcs._wcs.InconsistentAxisTypesError: ERROR 4 in wcs_types()
        #  Unmatched celestial axes."
        # For that reason we try a different approach:
        wv_cunit1, wv_crpix1, wv_crval1, wv_cdelt1 = get_wvparam_from_wcs3d(wcs3d)
        wcs2d = wcs.WCS(naxis=2)
        wcs2d.wcs.crpix = [wv_crpix1.value, 1]  # reference pixel coordinate
        wcs2d.wcs.crval = [wv_crval1.value, 0]  # world coordinate at reference pixel
        wcs2d.wcs.cdelt = [wv_cdelt1.value, 1]
        wcs2d.wcs.ctype = ["WAVE", ""]   # ToDo: fix this
        wcs2d.wcs.cunit = [wv_cunit1, u.pix]
        if bitpix == 16:
            # round to integer and save as BITPIX=16 (unsigned short)
            hdu = fits.PrimaryHDU(np.round(image2d_rss).astype(np.uint16))
        elif bitpix == -32:
            hdu = fits.PrimaryHDU(image2d_rss.astype(np.float32))
        else:
            raise ValueError(f'Unsupported BITPIX value: {bitpix}')
        pos0 = len(hdu.header) - 1
        hdu.header.extend(wcs_to_header_using_cd_keywords(wcs2d), update=True)
        hdu.header.update(header_keys)
        hdu.header.insert(
            pos0, ('COMMENT', "FITS (Flexible Image Transport System) format is defined in 'Astronomy"))
        hdu.header.insert(
            pos0 + 1, ('COMMENT', "and Astrophysics', volume 376, page 359; bibcode: 2001A&A...376..359H"))
        hdul = fits.HDUList([hdu])
        outfile = f'{prefix_intermediate_fits}_rss_2D_method{method}.fits'
        print(f'Saving file: {outfile}')
        hdul.writeto(outfile, overwrite='yes')



#
# Copyright 2024-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

from astropy.units import Unit
import pprint

from numina.tools.ctext import ctext

from .raise_valueerror import raise_ValueError

pp = pprint.PrettyPrinter(indent=1, sort_dicts=False)


def set_wavelength_unit_and_range(scene_fname, scene_block, wmin, wmax, verbose):
    """Set the wavelength unit and range for a scene block.

    Parameters
    ----------
    scene_fname : str
        YAML scene file name.
    scene_block : dict
        Dictonary storing the scene block.
    wmin : `~astropy.units.Quantity`
        Minimum wavelength covered by the detector.
    wmax : `~astropy.units.Quantity`
        Maximum wavelength covered by the detector.
    verbose : bool
       If True, display additional information.

    Returns
    -------
    wave_unit : `~astropy.units.core.Unit`
        Wavelength unit to be used in the scene block.
    wave_min : `~astropy.units.Quantity`
        Minimum wavelength to be used in the scene block.
    wave_max : `~astropy.units.Quantity`
        Maximum wavelength to be used in the scene block.

    """

    expected_keys_in_spectrum = {'type'}

    spectrum_keys = set(scene_block['spectrum'].keys())
    if not expected_keys_in_spectrum.issubset(spectrum_keys):
        print(ctext(f'ERROR while processing: {scene_fname}', fg='red'))
        print(ctext('expected keys..: ', fg='blue') + f'{expected_keys_in_spectrum}')
        print(ctext('keys found.....: ', fg='blue') + f'{spectrum_keys}')
        list_unexpected_keys = list(spectrum_keys.difference(expected_keys_in_spectrum))
        if len(list_unexpected_keys) > 0:
            print(ctext('unexpected keys: ', fg='red') + f'{list_unexpected_keys}')
        list_missing_keys = list(expected_keys_in_spectrum.difference(spectrum_keys))
        if len(list_missing_keys) > 0:
            print(ctext('missing keys:..: ', fg='red') + f'{list_missing_keys}')
        pp.pprint(scene_block)
        raise_ValueError(f'Invalid format in file: {scene_fname}')
    if 'wave_unit' in scene_block['spectrum']:
        wave_unit = scene_block['spectrum']['wave_unit']
    else:
        wave_unit = wmin.unit
        if verbose:
            print(ctext(f'Assuming wave_unit: {wave_unit}', faint=True))
    if 'wave_min' in scene_block['spectrum']:
        wave_min = float(scene_block['spectrum']['wave_min'])
    else:
        if verbose:
            print(ctext('Assuming wave_min: null', faint=True))
        wave_min = None
    if wave_min is None:
        wave_min = wmin.to(wave_unit)
    else:
        wave_min *= Unit(wave_unit)
    if 'wave_max' in scene_block['spectrum']:
        wave_max = float(scene_block['spectrum']['wave_max'])
    else:
        if verbose:
            print(ctext('Assuming wave_max: null', faint=True))
        wave_max = None
    if wave_max is None:
        wave_max = wmax.to(wave_unit)
    else:
        wave_max = wave_max * Unit(wave_unit)

    if verbose:
        print(ctext(f'{wave_min=}', faint=True))
        print(ctext(f'{wave_max=}', faint=True))
        print(ctext(f'{wave_unit=}', faint=True))

    return wave_unit, wave_min, wave_max




import argparse

import astropy.io.fits as fits


def insert_image(filename, extnum_filename, auximage, extnum_auximage):
    """Replace image in filename by another image (same size) in newimage.

    Parameters
    ----------
    filename : str
        File name where the new image will be inserted.
    extnum_filename : int
        Extension number in filename where the new image will be
        inserted. Note that the first extension is 1 (and not zero).
    auximage : str
        File name of the new image.
    extnum_auximage : int
        Extension number where the new image is located in auximage.
        Note that the first extension is 1 (and not zero).

    """

    # read the new image
    with fits.open(auximage) as hdulist:
        newimage = hdulist[extnum_auximage].data

    # open the destination image
    hdulist = fits.open(filename, mode='update')
    oldimage_shape = hdulist[extnum_filename].data.shape
    if oldimage_shape == newimage.shape:
        hdulist[extnum_filename].data = newimage
        hdulist.flush()
    else:
        print('filename shape:', oldimage_shape)
        print('newimage shape:', newimage.shape)
        print("ERROR: new image doesn't have the same shape")

    hdulist.close()


def main(args=None):
    parser = argparse.ArgumentParser()
    parser.add_argument("filename",
                        help="FITS file name")
    parser.add_argument("auximage",
                        help="image")
    parser.add_argument("--extnum_filename",
                        help="extension number (first=1)",
                        default=1, type=int)
    parser.add_argument("--extnum_auximage",
                        help="extension number in auximage (first=1)",
                        default=1, type=int)
    args = parser.parse_args(args=args)

    insert_image(args.filename, args.extnum_filename - 1,
                 args.auximage, args.extnum_auximage - 1)


if __name__ == '__main__':

    main()



import pytest

from ..assembly import assembly_instrument
from ..collection import load_paths_store


def load_from_tests(name, date_str):
    # pkg_paths = ["pattridrp.instrument.configs"]
    pkg_paths = ["numina.drps.tests.configs"]
    comp_store = load_paths_store(pkg_paths)
    ins = assembly_instrument(comp_store, name, date_str)
    return ins


def test_ins_load_fail():

    with pytest.raises(ValueError):
        load_from_tests("PATTRI", "2006-06-01T12:00:00.0")

    with pytest.raises(ValueError):
        load_from_tests("WRONG_NAME", "2016-06-01T12:00:00.0")



#
# Copyright 2019-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import enum
import typing

from typing_extensions import Self

if typing.TYPE_CHECKING:
    from numina.instrument.configorigin import ElementOrigin
    from numina.instrument.hwdevice import DeviceBase

from numina.instrument.components.wheel import Wheel


class FilterName(enum.Enum):
    U = 2131
    B = 2134
    V = 2135
    R = 2156
    I = 2158  # noqa
    Z = 2203


class PWheel(Wheel):
    """The particular wheel"""

    def __init__(
        self,
        cid,
        origin: "ElementOrigin | None" = None,
        parent: "DeviceBase | None" = None,
    ):
        super().__init__(cid, 5, origin=origin, parent=parent)

    @classmethod
    def from_component(
        cls,
        name: str,
        comp_id: str,
        origin: "ElementOrigin | None" = None,
        parent: "DeviceBase | None" = None,
        properties=None,
        setup=None,
    ) -> Self:
        def_filters = []
        if setup is not None:
            def_filters = setup.values["wheel"]["filters"]

        obj = cls.__new__(cls)
        obj.__init__(comp_id, origin=origin)

        for idx, el in enumerate(def_filters):
            obj.put_in_pos(FilterName[el], idx)

        return obj

    def configure_me_with_header(self, hdr):
        if "FILTER" in hdr:
            self.label = hdr["FILTER"]
        else:
            self.position = hdr.get("POS", 1)
        #
        self.is_configured = True

    def my_depends(self):
        return {"filter"}



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
"""Display the detected cosmic rays in the images."""
from astropy.table import Table
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.colors import ListedColormap, BoundaryNorm
import matplotlib.pyplot as plt
import numpy as np

import teareduce as tea


def display_detected_cr(num_images, image3d, median2d, median2d_corrected,
                        flag_integer_dilated, labels_cr, number_cr, mask_mediancr,
                        list_mask_single_exposures=None, mask_all_singles=None,
                        semiwindow=15, maxplots=-1, verify_cr=False,
                        color_scale='zscale', _logger=None):
    """Display the detected cosmic rays

    If list_mask_single_exposures is None, the plots correspond to the MEDIANCR
    computation. Otherwise, the plots correspond to problematic pixels (i.e.,
    cosmic-ray pixels flagged in all the single exposures).
    """
    if list_mask_single_exposures is None:
        output_basename = 'mediancr_identified'
        if mask_all_singles is not None:
            raise ValueError("mask_all_singles must be None when list_mask_single_exposures is None.")
    else:
        output_basename = 'problematic_pixels'
        if len(list_mask_single_exposures) != num_images:
            raise ValueError("len(list_mask_single_exposures) must be equal to num_images.")
        if verify_cr:
            raise ValueError("verify_cr cannot be True when list_mask_single_exposures is provided.")
        if mask_all_singles is None:
            raise ValueError("mask_all_singles must be provided when list_mask_single_exposures is provided.")

    # Determine the number of rows and columns for the plot,
    # considering that we want to plot also 3 additional images:
    # the median2d, the mask and the median2d_corrected
    num_plot_max = num_images
    if num_images == 3:
        nrows, ncols = 2, 3
        figsize = (10, 6)
    elif num_images in [4, 5]:
        nrows, ncols = 2, 4
        figsize = (13, 6)
    elif num_images == 6:
        nrows, ncols = 3, 3
        figsize = (13, 3)
    elif num_images in [7, 8, 9]:
        nrows, ncols = 3, 4
        figsize = (13, 9)
    else:
        _logger.warning("only the first 9 images will be plotted")
        nrows, ncols = 3, 4
        figsize = (13, 9)
        num_plot_max = 9
    pdf = PdfPages(f'{output_basename}.pdf')
    cr_table = Table(names=('CR_number', 'X_pixel', 'Y_pixel', 'Mask_value'), dtype=(int, int, int, int))
    cr_table_filename = f'{output_basename}.csv'
    maxplots_eff = maxplots
    if verify_cr:
        # In verify_cr mode, we plot all the cosmic rays
        maxplots_eff = number_cr
    elif maxplots_eff < 0:
        if number_cr > 200:
            maxplots_eff = 200
            _logger.warning(f"limiting to {maxplots_eff} plots (out of {number_cr} CRs detected)")
        else:
            maxplots_eff = number_cr
    _logger.info(f"generating {maxplots_eff} plots...")
    xlabel = 'X pixel (from 1 to NAXIS1)'
    ylabel = 'Y pixel (from 1 to NAXIS2)'
    naxis2, naxis1 = median2d.shape
    for i in range(min(number_cr, maxplots_eff)):
        ijloc = np.argwhere(labels_cr == i + 1)
        ic = int(np.mean(ijloc[:, 0]) + 0.5)
        jc = int(np.mean(ijloc[:, 1]) + 0.5)
        i1 = ic - semiwindow
        if i1 < 0:
            i1 = 0
        i2 = ic + semiwindow
        if i2 >= naxis2:
            i2 = naxis2 - 1
        j1 = jc - semiwindow
        if j1 < 0:
            j1 = 0
        j2 = jc + semiwindow
        if j2 >= naxis1:
            j2 = naxis1 - 1
        fig, axarr = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize)
        axarr = axarr.flatten()
        # Important: use interpolation=None instead of interpolation='None' to avoid
        # having blurred images when opening the PDF file with macos Preview
        cmap = 'viridis'
        cblabel = 'Number of counts'
        if color_scale == 'zscale':
            vmin, vmax = tea.zscale(median2d[i1:(i2+1), j1:(j2+1)])
        else:
            vmin = np.min(median2d[i1:(i2+1), j1:(j2+1)])
            vmax = np.max(median2d[i1:(i2+1), j1:(j2+1)])
        for k in range(num_plot_max):
            ax = axarr[k]
            title = title = f'image#{k+1}/{num_images}'
            tea.imshow(fig, ax, image3d[k][i1:(i2+1), j1:(j2+1)], vmin=vmin, vmax=vmax,
                       extent=[j1+0.5, j2+1.5, i1+0.5, i2+1.5],
                       xlabel=xlabel, ylabel=ylabel,
                       title=title, cmap=cmap, cblabel=cblabel, interpolation=None)
        if list_mask_single_exposures is None:
            # plots for mediancr
            for k in range(3):
                ax = axarr[k + num_plot_max]
                cmap = 'viridis'
                if k == 0:
                    image2d = median2d
                    title = 'median'
                    norm = None
                elif k == 1:
                    image2d = flag_integer_dilated
                    title = 'flag_integer_dilated'
                    # cmap = 'plasma'
                    color_list = ['black', 'grey', 'blue', 'red', 'yellow']
                    cmap = ListedColormap(color_list)
                    bounds = np.arange(-0.5, len(color_list), 1)   # integer limits: -0.5, 0.5, 1.5, ...
                    norm = BoundaryNorm(bounds, cmap.N)
                    cblabel = None
                elif k == 2:
                    image2d = median2d_corrected
                    title = 'median corrected'
                    norm = None
                else:
                    raise ValueError(f'Unexpected {k=}')
                if k in [0, 2]:
                    if color_scale == 'zscale':
                        vmin_, vmax_ = tea.zscale(image2d[i1:(i2+1), j1:(j2+1)])
                    else:
                        vmin_ = np.min(image2d[i1:(i2+1), j1:(j2+1)])
                        vmax_ = np.max(image2d[i1:(i2+1), j1:(j2+1)])
                else:
                    vmin_, vmax_ = None, None
                img, cax, cbar = tea.imshow(
                    fig, ax, image2d[i1:(i2+1), j1:(j2+1)], vmin=vmin_, vmax=vmax_,
                    extent=[j1+0.5, j2+1.5, i1+0.5, i2+1.5],
                    xlabel=xlabel, ylabel=ylabel,
                    title=title, cmap=cmap, norm=norm, cblabel=cblabel, interpolation=None
                )
                if k == 1:
                    cbar.set_ticks([0, 1, 2, 3, 4])
                    cbar.ax.yaxis.set_tick_params(length=0)
                    cbar.set_ticklabels(['0: no CR', '1: dilation', '2: mm', '3: la', '4: mm & la'])
                    # Overlay the grid of pixels
                    for idum in range(i1, i2 + 1):
                        ax.hlines(y=idum + 0.5, xmin=j1 + 0.5, xmax=j2 + 1.5, colors='w', lw=0.5, alpha=0.3)
                    for jdum in range(j1, j2 + 1):
                        ax.vlines(x=jdum + 0.5, ymin=i1 + 0.5, ymax=i2 + 1.5, colors='w', lw=0.5, alpha=0.3)
            nplot_missing = nrows * ncols - num_plot_max - 3
            if nplot_missing > 0:
                for k in range(nplot_missing):
                    ax = axarr[-k-1]
                    ax.axis('off')
            fig.suptitle(f'CR#{i+1}/{number_cr}')
        else:
            # print the coordinates of the problematic pixels
            ijloc = np.argwhere(labels_cr == i + 1)
            xproblematic = ijloc[:, 1] + 1  # FITS criterium
            yproblematic = ijloc[:, 0] + 1  # FITS criter
            # plots for problematic pixels
            for k in range(num_plot_max):
                ax = axarr[k + num_plot_max]
                title = title = f'mask#{k+1}/{num_images}'
                image2d = list_mask_single_exposures[k].astype(int)
                tea.imshow(fig, ax, image2d[i1:(i2+1), j1:(j2+1)], vmin=0, vmax=1,
                           extent=[j1+0.5, j2+1.5, i1+0.5, i2+1.5],
                           xlabel=xlabel, ylabel=ylabel,
                           title=title, cmap='plasma', cblabel='flag', interpolation=None)
                for xdum, ydum in zip(xproblematic, yproblematic):
                    ax.plot([xdum-0.5, xdum+0.5], [ydum-0.5, ydum+0.5], 'r-')
                    ax.plot([xdum-0.5, xdum+0.5], [ydum+0.5, ydum-0.5], 'r-')
            fig.suptitle(f'Problematic pixels, region #{i+1}/{number_cr}')

        plt.tight_layout()
        plt.show(block=False)

        if verify_cr:
            print('-' * 50)
        for idum in range(i1, i2 + 1):
            for jdum in range(j1, j2 + 1):
                if labels_cr[idum, jdum] == i + 1:
                    cr_table.add_row((i + 1, jdum + 1, idum + 1, flag_integer_dilated[idum, jdum]))
                    if verify_cr:
                        print(f'pixel (x,y) = ({jdum+1}, {idum+1}):  mask = {flag_integer_dilated[idum, jdum]}')
        if verify_cr:
            accept_cr = input(f"Accept this cosmic ray detection #{i+1} ([Y]es / [n]o / [a]ll / [s]top))? ")
            if accept_cr.lower() == 'n':
                _logger.info("removing cosmic ray detection #%d from the mask\n", i + 1)
                mask_mediancr[labels_cr == i + 1] = False
            elif accept_cr.lower() == 'a':
                verify_cr = False
                _logger.info("accepting all remaining cosmic ray detections\n")
                _logger.info("generating remaining plots without user interaction...")
            elif accept_cr.lower() == 's':
                _logger.info("stopping the program execution")
                raise SystemExit(0)
            else:
                _logger.info("keeping cosmic ray detection #%d in the mask", i + 1)
        pdf.savefig(fig, bbox_inches='tight')
        plt.close(fig)

    pdf.close()
    _logger.info("plot generation complete")
    _logger.info(f"saving {output_basename}.pdf file")
    cr_table.write(cr_table_filename, format='csv', overwrite=True)
    _logger.info("\n%s", cr_table)
    _logger.info("table of identified cosmic rays saved to %s", cr_table_filename)



#
# Copyright 2016-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

from .wheel import Wheel


class Stop:
    def __init__(self, name):
        self.name = name


class Open:
    def __init__(self, name):
        self.name = name


class Shutter(Wheel):
    """Shutter with two positions, OPEN and CLOSE"""

    def __init__(self, name="Shutter", parent=None):
        super().__init__(name, capacity=2, parent=parent)
        self.put_in_pos(Stop(name="STOP"), 0)  # FIXME
        self.put_in_pos(Open(name="OPEN"), 1)  # FIXME
        self.move_to(1)  # Open by default

    def configure_(self, value):
        # Let's see what is value:
        # a string
        if isinstance(value, str):
            val = value.lower()
            if val == "open":
                val = 1
            elif val == "closed":
                val = 0
            else:
                raise ValueError("Not allowed value %s", value)
        elif isinstance(value, int):
            val = value
        else:
            raise TypeError("Not allowed type %s", type(value))

        # Move to value
        self.move_to(val)

    def open(self):
        self.move_to(1)

    def close(self):
        self.move_to(0)



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
"""Combination of arrays avoiding coincident cosmic ray hits."""
import io
import logging
import os
import sys

import argparse
from astropy.io import fits
import numpy as np
from rich.logging import RichHandler
import yaml

from numina.tools.add_script_info_to_fits_history import add_script_info_to_fits_history
from numina.user.console import NuminaConsole
from numina._version import __version__

from .apply_crmasks import apply_crmasks
from .compute_crmasks import compute_crmasks
from .valid_parameters import VALID_COMBINATIONS


def main(args=None):
    """
    Main function to compute and apply CR masks.
    """

    parser = argparse.ArgumentParser(
        description="Combine 2D arrays using mediancr, meancrt or meancr methods."
    )

    parser.add_argument("inputyaml",
                        help="Input YAML file.",
                        type=str)
    parser.add_argument("--crmasks",
                        help="FITS file with cosmic ray masks",
                        type=str)
    parser.add_argument("--log-level",
                        help="Set the logging level",
                        type=str,
                        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
                        default='INFO')
    parser.add_argument("--record",
                        help="Record terminal output",
                        action="store_true")
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")

    args = parser.parse_args(args)

    if len(sys.argv) == 1:
        parser.print_usage()
        raise SystemExit()

    # Configure rich console
    console = NuminaConsole(record=args.record)

    if args.echo:
        console.print(f"[bright_red]Executing: {' '.join(sys.argv)}[/bright_red]\n", end='')

    # Configure logging
    if args.log_level in ['DEBUG', 'WARNING', 'ERROR', 'CRITICAL']:
        format_log = '%(name)s %(levelname)s %(message)s'
        handlers = [RichHandler(console=console, show_time=False, markup=True)]
    else:
        format_log = '%(message)s'
        handlers = [RichHandler(console=console, show_time=False, markup=True, show_path=False, show_level=False)]
    logging.basicConfig(
        level=args.log_level,
        format=format_log,
        handlers=handlers
    )
    logging.getLogger('matplotlib').setLevel(logging.ERROR)  # Suppress matplotlib debug logs

    # Welcome message
    console.rule("[bold magenta] Numina: Cosmic Ray Masks [/bold magenta]")

    # Display version info
    logger = logging.getLogger(__name__)
    logger.info(f"Using {__name__} version {__version__}")

    # Read parameters from YAML file
    with open(args.inputyaml, 'rt') as fstream:
        input_params = yaml.safe_load(fstream)
    logger.debug(f'{input_params=}')

    # Check that mandatory parameters are present
    if 'images' not in input_params:
        raise ValueError("'images' must be provided in input YAML file.")
    else:
        list_of_fits_files = input_params['images']
        if not isinstance(list_of_fits_files, list) or len(list_of_fits_files) < 3:
            raise ValueError("'images' must be a list of at least 3 FITS files.")
        for file in list_of_fits_files:
            if not os.path.isfile(file):
                raise FileNotFoundError(f"File {file} not found.")
            else:
                logger.info("found input file: %s", file)
    for item in ['gain', 'rnoise', 'bias']:
        if item not in input_params:
            raise ValueError(f"'{item}' must be provided in input YAML file.")

    # Default values for missing parameters in input YAML file
    if 'extnum' in input_params:
        extnum = int(input_params['extnum'])
    else:
        extnum = 0
        logger.info("extnum not provided, assuming extnum=0")

    # Read the input list of files, which should contain paths to 2D FITS files,
    # and load the arrays from the specified extension number.
    list_arrays = [fits.getdata(file, extnum=extnum) for file in input_params['images']]

    # Check if the list is empty
    if not list_arrays:
        raise ValueError("The input list is empty. Please provide a valid list of 2D arrays.")

    # Check that the requirements are provided
    if 'requirements' not in input_params:
        raise ValueError("'requirements' must be provided in input YAML file.")
    requirements = input_params['requirements']
    if not isinstance(requirements, dict):
        raise ValueError("'requirements' must be a dictionary.")
    if not requirements:
        raise ValueError("'requirements' dictionary is empty.")

    # Define parameters for compute_crmasks
    crmasks_params = dict()
    for key in ['gain', 'rnoise', 'bias']:
        crmasks_params[key] = input_params[key]
    for item in input_params['requirements']:
        crmasks_params[item] = input_params['requirements'][item]

    # If a FITS file with cosmic ray masks is provided, read it and skip
    # the computation of the masks. Otherwise, compute the masks.
    if args.crmasks is None:
        # Compute the different cosmic ray masks
        console.rule("[bold magenta] Computing cosmic ray masks [/bold magenta]")
        hdul_masks = compute_crmasks(
            list_arrays=list_arrays,
            _logger=logger,
            debug=(args.log_level == 'DEBUG'),
            **crmasks_params
        )
        # Save the cosmic ray masks to a FITS file
        output_masks = 'crmasks.fits'
        console.rule(f"Saving cosmic ray masks to [bold magenta]{output_masks}[/bold magenta]")
        hdul_masks.writeto(output_masks, overwrite=True)
        logger.info("Cosmic ray masks saved")
        buffer = io.StringIO()
        fits.info(output_masks, output=buffer)
        logger.info("%s", buffer.getvalue())
    else:
        if not os.path.isfile(args.crmasks):
            raise FileNotFoundError(f"File {args.crmasks} not found.")
        else:
            console.rule(f"reading cosmic ray masks from [bold magenta]{args.crmasks}[/bold magenta]")
            hdul_masks = fits.open(args.crmasks)
            buffer = io.StringIO()
            fits.info(args.crmasks, output=buffer)
            logger.info("%s", buffer.getvalue())

    # Apply cosmic ray masks
    for combination in VALID_COMBINATIONS:
        console.rule(f"Applying cosmic ray masks using [bold magenta]{combination}[/bold magenta]")
        output_combined = f'combined_{combination}.fits'
        combined, variance, maparray = apply_crmasks(
            list_arrays=list_arrays,
            hdul_masks=hdul_masks,
            combination=combination,
            use_lamedian=crmasks_params.get('use_lamedian', False),
            dtype=np.float32,
            apply_flux_factor=True,
            bias=input_params['bias'],
        )
        # Save the combined array, variance, and map to a FITS file
        logger.info("Saving combined (bias subtracted) array, variance, and map to [bold magenta]%s[/bold magenta]",
                    output_combined)
        hdu_combined = fits.PrimaryHDU(combined.astype(np.float32))
        add_script_info_to_fits_history(hdu_combined.header, args)
        hdu_combined.header.add_history('Contents of --inputlist:')
        for item in list_of_fits_files:
            hdu_combined.header.add_history(f'- {item}')
        hdu_combined.header.add_history(f"Masks UUID: {hdul_masks[0].header['UUID']}")
        hdu_variance = fits.ImageHDU(variance.astype(np.float32), name='VARIANCE')
        hdu_map = fits.ImageHDU(maparray.astype(np.int16), name='MAP')
        hdul = fits.HDUList([hdu_combined, hdu_variance, hdu_map])
        hdul.writeto(output_combined, overwrite=True)
        logger.info("Combined (bias subtracted) array, variance, and map saved")

    # Goodbye message
    console.rule("[bold magenta] Goodbye! [/bold magenta]")

    if args.record:
        log_filename = 'terminal_output.txt'
        with open(log_filename, 'wt') as f:
            f.write(console.export_text(styles=True))
        logger.info("terminal output recorded in %s", log_filename)


if __name__ == "__main__":

    main()




import numpy
from ..mode import mode_half_sample, mode_sex


def test_mode_halfsample():
    numpy.random.seed(1392838)
    a = numpy.random.normal(1000, 200, size=1000)
    a[:100] = numpy.random.normal(2000, 300, size=100)
    b = numpy.sort(a)
    m = mode_half_sample(b, is_sorted=True)
    assert numpy.allclose(m, 1041.9327885039545)


def test_mode_sextractor():
    numpy.random.seed(1392838)
    a = numpy.random.normal(1000, 200, size=1000)
    a[:100] = numpy.random.normal(2000, 300, size=100)
    b = numpy.sort(a)
    m = mode_sex(b)
    assert numpy.allclose(m, 912.78563284477332)



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Unit test for extract"""

import numpy as np
from numpy.testing import assert_allclose

from ..extract import extract_simple_rss_apers
from ..extract import Aperture
from ..traces import axis_to_dispaxis


def test_extract_simple_rss_apers_flat():

    img = np.ones((89, 100))

    bbox = [5, 71, 6, 64]
    # Check with flat boundaries
    # axis = 0
    aa1 = 2.0
    aa2 = 5.5
    axis = 0
    dispaxis = axis_to_dispaxis(axis)
    def b1(x): return aa1 * np.ones_like(x)  # noqa: E731
    def b2(x): return aa2 * np.ones_like(x)  # noqa: E731

    result = np.zeros((1, img.shape[1]))
    result[0, bbox[0]:bbox[1]+1] = aa2 - aa1

    aper = Aperture(bbox, [b1, b2], axis=0)
    out = extract_simple_rss_apers(img, [aper], axis=axis)

    assert out.shape[1] == img.shape[dispaxis]
    assert_allclose(out, result)

    # Check with flat boundaries
    # axis = 1
    aa3 = 1.9
    aa4 = 5.5
    axis = 1
    dispaxis = axis_to_dispaxis(axis)
    def b3(x): return aa3 * np.ones_like(x)  # noqa: E731
    def b4(x): return aa4 * np.ones_like(x)  # noqa: E731

    result = np.zeros((1, img.shape[0]))
    result[0, bbox[2]:bbox[3]+1] = aa4 - aa3

    aper = Aperture(bbox, [b3, b4], axis=1)
    out = extract_simple_rss_apers(img, [aper], axis=axis)
    assert out.shape[1] == img.shape[dispaxis]

    assert_allclose(out, result)


def test_extract_simple_rss_apers_line():

    img = np.ones((89, 100))
    bbox = [5, 71, 6, 64]

    # Check with flat boundaries
    # axis = 0
    aa1 = 2.0
    aa2 = 5.5

    axis = 0
    dispaxis = axis_to_dispaxis(axis)
    def b1(x): return aa1 + 1.1 * (x - 50) / 100  # noqa: E731
    def b2(x): return aa2 + 1.1 * (x - 50) / 100  # noqa: E731

    result = np.zeros((1, img.shape[1]))
    result[0, bbox[0]:bbox[1]+1] = aa2 - aa1

    aper = Aperture(bbox, [b1, b2], axis=0)
    out = extract_simple_rss_apers(img, [aper], axis=axis)

    assert out.shape[1] == img.shape[dispaxis]
    assert_allclose(out, result)

    # Check with flat boundaries
    # axis = 1
    aa3 = 1.9
    aa4 = 5.5

    def b3(x): return aa3 + 1.2 * (x - 44) / 89  # noqa: E731
    def b4(x): return aa4 + 1.2 * (x - 44) / 89  # noqa: E731
    axis = 1
    dispaxis = axis_to_dispaxis(axis)

    result = np.zeros((1, img.shape[0]))
    result[0, bbox[2]:bbox[3]+1] = aa4 - aa3

    aper = Aperture(bbox, [b3, b4], axis=1)
    out = extract_simple_rss_apers(img, [aper], axis=axis)

    assert out.shape[1] == img.shape[dispaxis]
    assert_allclose(out, result)



#
# Copyright 2016-2020 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import numpy


def detrend(arr, x=None, deg=5, tol=1e-3, maxloop=10):
    """
    Compute a baseline trend of a signal

    Parameters
    ----------
    arr
    x
    deg
    tol
    maxloop

    Returns
    -------

    """

    xx = numpy.arange(len(arr)) if x is None else x
    base = arr.copy()
    trend = base
    pol = numpy.ones((deg + 1,))
    for _ in range(maxloop):
        pol_new = numpy.polyfit(xx, base, deg)
        pol_norm = numpy.linalg.norm(pol)
        diff_pol_norm = numpy.linalg.norm(pol - pol_new)
        if diff_pol_norm / pol_norm < tol:
            break
        pol = pol_new
        trend = numpy.polyval(pol, xx)
        base = numpy.minimum(base, trend)
    return trend



#
# Copyright 2008-2014 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""
Basic Data Products
"""

from astropy.io import fits


class DataFrame(object):
    """A handle to a image in disk or in memory."""

    def __init__(self, frame=None, filename=None):
        if frame is None and filename is None:
            raise ValueError('only one in frame and filename can be None')
        self.frame = frame
        self.filename = filename

    def open(self):
        if self.frame is None:
            return fits.open(self.filename, mode='readonly')
        else:
            return self.frame

    @property
    def label(self):
        return self.filename

    def __repr__(self):
        if self.frame is None:
            return f"DataFrame(filename={self.filename!r})"
        elif self.filename is None:
            return f"DataFrame(frame={self.frame!r})"
        else:
            fmt = "DataFrame(filename=%r, frame=%r)"
            return fmt % (self.filename, self.frame)

    def __numina_load__(self, obj):
        if obj is None:
            return None
        else:
            return DataFrame(filename=obj)



#
# Copyright 2014-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import os
import tarfile
import warnings
import sys

import pytest

if "pytest_benchmark" in sys.modules:
    HAS_BENCHMARCK = True
else:
    from .nobenchmark import benchmark  # noqa: F401
    HAS_BENCHMARCK = False


import numina.util.context as ctx
from .drpmocker import DRPMocker
from .testcache import download_cache
from .pytest_resultcmp import ResultCompPlugin


@pytest.fixture
def numinatmpdir(tmpdir):
    """Return a temporary directory for recipe testing"""

    tmpdir.mkdir('_work')
    tmpdir.mkdir('_data')
    return tmpdir


@pytest.fixture
def numinatpldir(tmpdir, request):
    """Return a temporary dataset for recipe testing.

    Return a temporary directory path object
    for numina, where a dataset has been downloaded
    from a remote location, based on
    the module variable BASE_URL and the test function name
    """

    # Name of the dataset based on the function name
    tarname = request.function.__name__[5:]
    # Base url to donwload
    base = getattr(request.module, 'BASE_URL')
    url = base + tarname + '.tar.gz'

    downloaded = download_cache(url)

    tmpdir.chdir()

    # Uncompress
    with tarfile.open(downloaded.name, mode="r:gz") as tar:
        tar.extractall()

    os.remove(downloaded.name)
    os.chdir('tpl')
    return tmpdir


@pytest.fixture
def drpmocker(monkeypatch):
    """A fixture that mocks the loading of DRPs"""
    return DRPMocker(monkeypatch)


@pytest.fixture(scope='module')
def datamanager_remote(tmp_path_factory, request):
    """Return a DataManager object create from a remote dataset"""
    from numina.user.helpers import create_datamanager

    req_base_default = "https://guaix.fis.ucm.es/data/"
    req_base = getattr(request.module, 'TEST_SET_HOST', req_base_default)
    req_tarname = getattr(request.module, 'TEST_SET_FILE')
    req_datadir = getattr(request.module, 'TEST_SET_DATADIR', 'data')
    req_control = getattr(
        request.module, 'TEST_SET_CONTROL', "control_v2.yaml")

    basedir = tmp_path_factory.mktemp('manager')

    datadir = basedir / req_datadir  # pathlib syntax
    reqfile = basedir / req_control

    if req_tarname is None:
        raise ValueError('Undefined TEST_SET_FILE')

    url = req_base + req_tarname

    # Download everything
    with ctx.working_directory(basedir):

        downloaded = download_cache(url)

        # Uncompress
        with tarfile.open(downloaded.name, mode="r:gz") as tar:
            tar.extractall()

        os.remove(downloaded.name)

    # Insert OBS in the control file....
    config = {'tool.run': {}}
    section = config['tool.run']
    section['basedir'] = basedir
    section['datadir'] = datadir
    dm = create_datamanager(config, reqfile)

    # This is not really needed...
    # If everything is in the file already
    # with working_directory(basedir):
    #     obsresults = ['obs_ids.yaml']
    #     sessions, loaded_obs = load_observations(obsresults, is_session=False)
    #    dm.backend.add_obs(loaded_obs)

    return dm


def pytest_report_header(config):
    if not HAS_BENCHMARCK:
        return "pytest-benchmark not installed"
    return ""


def pytest_addoption(parser):
    parser.addoption(
        '--resultcmp', action='store_true',
        help="enable comparison of recipe results to reference results stored"
    )
    parser.addoption(
        '--resultcmp-generate-path',
        help="directory to generate reference files in, relative to location where py.test is run",
        action='store'
    )
    parser.addoption(
        '--resultcmp-reference-path',
        help="directory containing reference files, relative to location where py.test is run",
        action='store'
    )


def pytest_configure(config):

    config.getini('markers').append(
        'result_compare: Apply to tests that provide recipe results to compare with a reference')

    if config.getoption("--resultcmp", default=False) \
            or config.getoption("--resultcmp-generate-path", default=None) is not None:

        reference_dir = config.getoption("--resultcmp-reference-path")
        generate_dir = config.getoption("--resultcmp-generate-path")

        if reference_dir is not None and generate_dir is not None:
            warnings.warn(
                "Ignoring --resultcmp-reference-path since --resultcmp-generate-path is set")

        if reference_dir is not None:
            reference_dir = os.path.abspath(reference_dir)
        if generate_dir is not None:
            reference_dir = os.path.abspath(generate_dir)

        # default_format = config.getoption("--resultcmp-default-format") or 'text'
        config.pluginmanager.register(ResultCompPlugin(
            config, reference_dir=reference_dir, generate_dir=generate_dir
        ))



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
"""Estimate the limits for the diagnostic plot."""
import numpy as np


def estimate_diagnostic_limits(rng, gain, rnoise, maxvalue, num_images, npixels):
    """Estimate the limits for the diagnostic plot.

    Parameters
    ----------
    rng : numpy.random.Generator
        Random number generator for reproducibility.
    gain : float
        Gain value (in e/ADU) of the detector.
    rnoise : float
        Readout noise (in ADU) of the detector.
    maxvalue : float
        Maximum pixel value (in ADU) of the detector after
        subtracting the bias.
    num_images : int
        Number of different exposures.
    npixels : int
        Number of simulations to perform for each corner of the
        diagnostic plot.
    """

    if maxvalue < 0:
        maxvalue = 0.0
    xdiag_min = np.zeros(npixels, dtype=float)
    xdiag_max = np.zeros(npixels, dtype=float)
    ydiag_min = np.zeros(npixels, dtype=float)
    ydiag_max = np.zeros(npixels, dtype=float)
    for i in range(npixels):
        # lower limits
        data = rng.normal(loc=0, scale=rnoise, size=num_images)
        min1d = np.min(data)
        median1d = np.median(data)
        xdiag_min[i] = median1d
        ydiag_min[i] = median1d - min1d
        # upper limits
        lam = np.array([maxvalue] * num_images)
        data = rng.poisson(lam=lam * gain).astype(float) / gain
        if rnoise > 0:
            data += rng.normal(loc=0, scale=rnoise, size=num_images)
        min1d = np.min(data)
        median1d = np.median(data)
        xdiag_max[i] = median1d
        ydiag_max[i] = median1d - min1d
    xdiag_min = np.min(xdiag_min)
    ydiag_min = np.min(ydiag_min)
    xdiag_max = np.max(xdiag_max)
    ydiag_max = np.max(ydiag_max)
    return xdiag_min, xdiag_max, ydiag_min, ydiag_max



#
# Copyright 2019-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import itertools
import json
import os
import pathlib
import typing

if typing.TYPE_CHECKING:
    from collections.abc import Iterable


import attrs

import importlib_resources

from .configorigin import ElementOrigin


FileLike = str | os.PathLike


@attrs.define
class ComponentCollection:
    dirname = attrs.field()
    paths = attrs.field()


def load_paths_store(
    pkg_paths: "Iterable[FileLike] | None" = None,
    file_paths: "Iterable[FileLike] | None" = None,
) -> dict[str, typing.Any]:
    comp_store = {}
    # Prepare file paths
    if file_paths is None:
        file_paths = []
    if pkg_paths is None:
        pkg_paths = []

    paths1 = [pathlib.Path(f_path) for f_path in file_paths]
    paths2 = [importlib_resources.files(p_path) for p_path in pkg_paths]

    for path in itertools.chain(paths1, paths2):
        for obj in path.iterdir():
            if obj.suffix == ".json":
                with open(obj) as fd:
                    cont = json.load(fd)
                    cont["origin"] = ElementOrigin.from_dict(cont)
                    comp_store[obj.name] = cont

    return comp_store




import pytest


def test_warns_qc():

    with pytest.warns(DeprecationWarning):
        import numina.core.qc  # noqa: F401


def test_warns_products():

    with pytest.warns(DeprecationWarning):
        import numina.core.products  # noqa: F401



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""DAL for dictionary-based database of products."""

import os
import logging
import json
import operator

import numina.store
from numina.exceptions import NoResultFound
from numina.util.fqn import fully_qualified_name
from numina.util.context import working_directory

from .dictdal import BaseHybridDAL
from .stored import StoredProduct, StoredResult


_logger = logging.getLogger(__name__)


class BackendTable(object):
    def __init__(self):
        self.table_contents = {}
        self.table_index = []

    def new_id(self):
        if self.table_index:
            newidx = self.table_index[-1] + 1
        else:
            newidx = 1
        self.table_index.append(newidx)
        return newidx

    def insert(self, prodid, prod_reg):
        self.table_contents[prodid] = prod_reg


class Backend(BaseHybridDAL):

    def __init__(self, drps, base, extra_data=None, basedir=None, components=None,
                 filename=None):

        self.filename = filename
        temp_ob_ids = []
        obtable = base.get('oblocks', {})
        if isinstance(obtable, dict):
            for ob in obtable:
                temp_ob_ids.append(ob)
            obdict = obtable
        elif isinstance(obtable, list):
            obdict = {}
            for ob in obtable:
                temp_ob_ids.append(ob['id'])
                obdict[ob['id']] = ob
        else:
            raise TypeError("'oblocks' must be a list or a dictionary")

        self.db_tables = {}

        self.db_tables['tasks'] = base.get('tasks', {})
        self.db_tables['tasks_index'] = base.get('tasks_index', [])

        self.db_tables['results'] = base.get('results', {})
        self.db_tables['results_index'] = base.get('results_index', [])

        self.db_tables['products'] = base.get('products', {})
        self.db_tables['products_index'] = base.get('products_index', [])

        super(Backend, self).__init__(
            drps, obdict, base,
            extra_data=extra_data,
            basedir=basedir,
            components=components
        )

        self.ob_ids = temp_ob_ids
        self.db_tables['oblocks'] = self.ob_table
        self.db_tables['requirements'] = self.req_table
        self.db_tables['oblocks_index'] = self.ob_ids
        self.prod_table = self.db_tables['products']

    def dump_data(self):
        state = {}
        state['version'] = 2
        database = {}
        state['database'] = database
        for name, val in self.db_tables.items():
            database[name] = val
        return state

    def dump_to_file(self):
        if self.filename:
            with open(self.filename, 'w') as fp:
                self.dump(fp)

    def new_id(self, table_index):
        if table_index:
            newidx = table_index[-1] + 1
        else:
            newidx = 1
        table_index.append(newidx)
        self.dump_to_file()
        return newidx

    def new_task_id(self, request, request_params):
        newidx = self.new_id(self.db_tables['tasks_index'])
        return newidx

    def new_result_id(self):
        newidx = self.new_id(self.db_tables['results_index'])
        return newidx

    def new_product_id(self):
        newidx = self.new_id(self.db_tables['products_index'])
        return newidx

    def new_task(self, request, request_params):

        task = super(Backend, self).new_task(request, request_params)

        task_reg = {
            'id': task.id, 'state': task.state,
            'time_create': task.time_create.strftime('%FT%T'),
            'request': request,
            'request_params': request_params,
            'request_runinfo': task.request_runinfo
        }
        _logger.debug('insert task=%d in backend', task.id)
        self.db_tables['tasks'][task.id] = task_reg
        self.dump_to_file()
        return task

    def update_task(self, task):
        _logger.debug('update task=%d in backend', task.id)
        task_reg = self.db_tables['tasks'][task.id]
        task_reg['state'] = task.state
        task_reg['time_start'] = task.time_start.strftime('%FT%T')
        task_reg['time_end'] = task.time_end.strftime('%FT%T')
        task_reg['request'] = task.request
        task_reg['request_params'] = task.request_params
        task_reg['request_runinfo'] = task.request_runinfo
        self.dump_to_file()

    def update_result(self, task, serialized, filename):
        _logger.debug('update result of task=%d in backend', task.id)
        newix = self.new_result_id()
        _logger.debug('result_id=%d in backend', newix)
        result = task.result
        if result is None:
            return

        res_dir = task.request_runinfo['results_dir']
        result_reg = {
            'id': newix,
            'task_id': task.id,
            'uuid': str(task.result.uuid),
            'qc': task.result.qc.name,
            'mode': task.request_runinfo['mode'],
            'instrument': task.request_runinfo['instrument'],
            'pipeline': task.request_runinfo['pipeline'],
            'time_create': task.time_end.strftime('%FT%T'),
            'time_obs': '',
            'recipe_class': task.request_runinfo['recipe_class'],
            'recipe_fqn': task.request_runinfo['recipe_fqn'],
            'oblock_id': task.request_params['oblock_id'],
            'result_dir': res_dir,
            'result_file': filename
        }
        self.db_tables['results'][newix] = result_reg

        for key, prod in result.stored().items():

            DB_PRODUCT_KEYS = [
                'instrument',
                'observation_date',
                'uuid',
                'quality_control'
            ]

            if prod.type.isproduct():
                newprod = self.new_product_id()
                _logger.debug('product_id=%d in backend', newprod)
                internal_value = getattr(result, key)
                ometa = prod.type.extract_db_info(
                    internal_value, DB_PRODUCT_KEYS)
                prod_reg = {
                    'id': newprod,
                    'result_id': newix,
                    'qc': task.result.qc.name,
                    'instrument': task.request_runinfo['instrument'],
                    'time_create': task.time_end.strftime('%FT%T'),
                    'time_obs': ometa['observation_date'].strftime('%FT%T'),
                    'tags': ometa['tags'],
                    'uuid': ometa['uuid'],
                    'oblock_id': task.request_params['oblock_id'],
                    'type': prod.type.name(),
                    'type_fqn': fully_qualified_name(prod.type),
                    'content': os.path.join(res_dir, serialized['values'][key])
                }
                self.db_tables['products'][newprod] = prod_reg
        self.dump_to_file()

    def _search_prod_table(self, name, tipo, obsres):
        """Returns the first coincidence..."""

        instrument = obsres.instrument

        # conf = str(obsres.configuration.origin.uuid)

        drp = self.drps.query_by_name(instrument)
        label = drp.product_label(tipo)
        # Strip () is present
        if label.endswith("()"):
            label_alt = label[:-2]
        else:
            label_alt = label

        # search results of these OBs
        if isinstance(self.prod_table, dict):
            iter_over = self.prod_table.values()
        else:
            iter_over = self.prod_table
        pred_func = tipo.query_expr.fill_placeholders(**obsres.tags)
        for prod in iter_over:
            pi = prod['instrument']
            pk = prod['type']
            pt = prod['tags']

            if pi == instrument and ((pk == label) or (pk == label_alt)):
                pred_val = pred_func.eval(**pt)
                # pred_val must be boolean to check if it is valid
                # if the expression is incomplete it may be non-bool
                if pred_val:
                    # this is a valid product
                    # We have found the result, no more checks
                    # Make a copy
                    rprod = dict(prod)
                    path = prod['content']
                    # Check if path is absolute
                    if not os.path.isabs(path):
                        path = os.path.join(self.basedir, path)
                    rprod['content'] = numina.store.load(tipo, path)
                    return StoredProduct(**rprod)
        else:
            msg = f'type {label_alt} compatible with tags {obsres.tags} not found'
            raise NoResultFound(msg)

    def search_result_id(self, node_id, tipo, field, mode=None):
        cobsres = self.obsres_from_oblock_id(node_id)

        if mode is not None:
            # mode must match
            if cobsres.mode != mode:
                msg = f"requested mode '{mode}' and obsmode '{cobsres.mode}' do not match"
                raise NoResultFound(msg)

        try:
            candidates = []
            for idx, val in self.db_tables['results'].items():
                if node_id == val['oblock_id']:
                    candidates.append(val)

            s_can = sorted(candidates, key=operator.itemgetter(
                'time_create'), reverse=True)
            if s_can:
                result_reg = s_can[0]
                directory = result_reg.get('result_dir', '')
                filename = result_reg['result_file']
                # change directory to open result file
                with working_directory(os.path.join(self.basedir, directory)):
                    with open(filename) as fd:
                        data = json.load(fd)
                        stored_result = StoredResult.load_data(data)
                try:
                    content = getattr(stored_result, field)
                except AttributeError:
                    raise NoResultFound(f'no field {field} found in result')

                st = StoredProduct(
                    id=result_reg['id'],
                    content=content,
                    tags={}
                )
                return st
            else:
                msg = f"result of mode '{field}' id={node_id} not found"
                raise NoResultFound(msg)
        except KeyError as err:
            msg = f"field '{field}' not found in result of mode '{cobsres.mode}' id={node_id}"
            raise NoResultFound(msg) from err

    def search_session_ids(self):
        for obs_id in self.ob_ids:
            obdict = self.ob_table[obs_id]
            enabled = obdict.get('enabled', True)
            if (not enabled) or (obdict['mode'] in self._RESERVED_MODE_NAMES):
                # ignore these OBs
                continue

            yield obs_id

    def build_recipe_result(self, result_id):
        result_reg = self.db_tables['results'][result_id]
        result_file = result_reg['result_file']
        result_dir = result_reg.get('result_dir', '')

        with working_directory(os.path.join(self.basedir, result_dir)):
            with open(result_file) as fd:
                import json
                data = json.load(fd)
                return StoredResult.load_data(data)



#

from typing import TypeVar, Type, TYPE_CHECKING

if TYPE_CHECKING:
    from .configorigin import ElementOrigin
    from .property import PropertyBase
    from typing_extensions import Self

from .device import DeviceBase


class ComponentGeneric(DeviceBase):
    """Class representing a device component from a file"""

    _property_names = []
    _property_readonly = []
    _property_configurable = []
    _mappings = {}

    def __init__(
        self,
        name: str,
        origin: "ElementOrigin | None" = None,
        parent: DeviceBase | None = None,
    ):
        super().__init__(name, origin=origin, parent=parent)
        # This attribute is used to store the values of the properties
        self._internal_state = {}

        for prop_name in self._property_names:
            deflts = self.__class__.__dict__[prop_name].defaults()

            for dk, dv in deflts.items():
                self._internal_state[dk] = dv

    @classmethod
    def from_component(
        cls,
        name: str,
        comp_id: str,
        origin: "ElementOrigin | None" = None,
        parent: DeviceBase | None = None,
        properties=None,
        setup=None,
    ) -> "Self":
        class_name = "{base_name}_{name}".format(base_name=cls.__name__, name=name)
        ncls = make_component_class(class_name, cls, properties=properties)
        obj = ncls.__new__(ncls)
        obj.__init__(comp_id, origin, parent)
        return obj

    def get_value(self, path: str, **state):
        device_path, prop_name = self.split_path(path)

        dev = self.get_device_seq(device_path)
        # save state and restore state
        current_state = dev.config_info()
        dev.configure_me(state)
        value = getattr(dev, prop_name)
        dev.configure(current_state)
        return value

    def configure_me_with_header(self, hdr):
        for el, key in self._mappings.items():
            val = key.get_value(hdr)
            if val is not None:
                setattr(self, el, val)

    def my_depends(self) -> set[str]:
        """Compute the dependencies for me"""
        my_deps = set([])
        for prop_name in self._property_names:
            depends = self.__class__.__dict__[prop_name].depends
            my_deps.update(depends)
        return my_deps

    def get_property_names(self):
        return self._property_configurable


class InstrumentGeneric(ComponentGeneric):
    pass


CG = TypeVar("CG", bound=ComponentGeneric)


def make_component_class(
    name: str,
    baseclass: Type[ComponentGeneric],
    properties: "dict[str, PropertyBase] | None" = None,
) -> Type[ComponentGeneric]:
    # Inject properties as descriptors in the class instance
    cls = type(name, (baseclass,), {})

    cls._property_readonly = []
    cls._property_configurable = []
    cls._property_names = []
    cls._mappings = {}

    if properties is not None:
        cls._property_names = list(properties.keys())
        for p_name, prop in properties.items():
            if prop.is_readonly:
                cls._property_readonly.append(p_name)
            else:
                cls._property_configurable.append(p_name)
            if prop.mapping is not None:
                cls._mappings[p_name] = prop.mapping
            # This method is not called automatically here
            prop.__set_name__(cls, p_name)
            setattr(cls, p_name, prop)
    return cls




from ..arccalibration import SolutionArcCalibration, WavecalFeature, CrLinear

orig = {'cr_linear': {'cdelt': 3.0, 'crmax': 4600.0, 'crmin': 2300.0,
                      'crpix': 1200.0,
                      'crval': 12.0},
        'features': [
            {'category': 'A', 'xpos': 100.0, 'reference': 3210.0,
             'wavelength': 3100.0, 'line_ok': True, 'ypos': 0.0, 'peak': 0.0,
             'funcost': 12.0, 'fwhm': 0.0,
             'lineid': 1},
            {'category': 'A', 'xpos': 150.0, 'reference': 3310.0,
             'wavelength': 3150.0, 'line_ok': True, 'ypos': 0.0, 'peak': 0.0,
             'funcost': 12.0, 'fwhm': 0.0,
             'lineid': 2},
            {'category': 'C', 'xpos': 250.0, 'reference': 3410.0,
             'wavelength': 3250.0, 'line_ok': True, 'ypos': 0.0, 'peak': 0.0,
             'funcost': 13.0, 'fwhm': 0.0,
             'lineid': 3}], 'coeff': [3000.0, 1.0], 'residual_std': 1.0}


def create_solution(orig):

    # Create Features
    features = []
    for feature in orig['features']:
        m = WavecalFeature(**feature)
        features.append(m)

    cr_linear = CrLinear(**orig['cr_linear'])
    mm = SolutionArcCalibration(
        features, orig['coeff'], orig['residual_std'], cr_linear)
    return mm


def test_serialize1():

    mm = create_solution(orig)
    assert orig == mm.__getstate__()


def test_serialize2():

    mm = create_solution(orig)

    reconstructed = SolutionArcCalibration.__new__(SolutionArcCalibration)

    reconstructed.__setstate__(orig)

    assert reconstructed == mm
